"""
Test suite for DelegateParameter
"""
from typing import cast

import hypothesis.strategies as hst
import pytest
from hypothesis import given

from qcodes.instrument.parameter import DelegateParameter, Parameter, ParamRawDataType
from .conftest import BetterGettableParam

# Disable warning that is created by using fixtures
# pylint: disable=redefined-outer-name


@pytest.fixture()
def numeric_val():
    yield 1


@pytest.fixture()
def simple_param(numeric_val):
    yield Parameter('testparam', set_cmd=None, get_cmd=None,
                    scale=2, offset=17,
                    label='Test Parameter', unit='V',
                    initial_value=numeric_val)


@pytest.fixture(params=[True, False])
def make_observable_parameter(request):
    def make_parameter(*args, override_getset: bool = True, **kwargs):
        class ObservableParam(Parameter):
            def __init__(self, *args, **kwargs):
                self.instr_val = None
                super().__init__(*args, **kwargs)

            def set_raw(  # pylint: disable=method-hidden
                    self, value: ParamRawDataType) -> None:
                self.instr_val = value

            def get_raw(  # pylint: disable=method-hidden
                    self) -> ParamRawDataType:
                return self.instr_val

            def get_instr_val(self):
                return self.instr_val

        if request.param:
            if not override_getset:
                pytest.skip()
            param = ObservableParam(*args, **kwargs)
        else:
            val = None

            def set_cmd(value):
                nonlocal val
                val = value

            def get_cmd():
                nonlocal val
                return val

            p = Parameter(*args, **kwargs,  # type: ignore[misc]
                          set_cmd=set_cmd, get_cmd=get_cmd)
            param = cast(ObservableParam, p)
            param.get_instr_val = get_cmd  # type: ignore[assignment]
        return param
    yield make_parameter


def test_observable_parameter(make_observable_parameter, numeric_val):
    p = make_observable_parameter('testparam')
    p(numeric_val)
    assert p.get_instr_val() == numeric_val


def test_observable_parameter_initial_value(make_observable_parameter,
                                            numeric_val):
    t = make_observable_parameter(
        'observable_parameter', initial_value=numeric_val)
    assert t.get_instr_val() == numeric_val


def test_same_value(simple_param):
    d = DelegateParameter('test_delegate_parameter', simple_param)
    assert d() == simple_param()


def test_same_label_and_unit_on_init(simple_param):
    """
    Test that the label and unit get used from source parameter if not
    specified otherwise.
    """
    d = DelegateParameter('test_delegate_parameter', simple_param)
    assert d.label == simple_param.label
    assert d.unit == simple_param.unit


def test_overwritten_unit_on_init(simple_param):
    d = DelegateParameter('test_delegate_parameter', simple_param, unit='Ohm')
    assert d.label == simple_param.label
    assert not d.unit == simple_param.unit
    assert d.unit == 'Ohm'


def test_overwritten_label_on_init(simple_param):
    d = DelegateParameter('test_delegate_parameter', simple_param,
                          label='Physical parameter')
    assert d.unit == simple_param.unit
    assert not d.label == simple_param.label
    assert d.label == 'Physical parameter'


def test_get_set_raises(simple_param):
    """
    Test that providing a get/set_cmd kwarg raises an error.
    """
    for kwargs in ({'set_cmd': None}, {'get_cmd': None}):
        with pytest.raises(KeyError) as e:
            DelegateParameter('test_delegate_parameter', simple_param, **kwargs)
        assert str(e.value).startswith('\'It is not allowed to set')


def test_scaling(simple_param, numeric_val):
    scale = 5
    offset = 3
    d = DelegateParameter(
        'test_delegate_parameter', simple_param, offset=offset, scale=scale)

    simple_param(numeric_val)
    assert d() == (numeric_val - offset) / scale

    d(numeric_val)
    assert simple_param() == numeric_val * scale + offset


def test_scaling_delegate_initial_value(simple_param, numeric_val):
    scale = 5
    offset = 3
    DelegateParameter(
        'test_delegate_parameter', simple_param, offset=offset, scale=scale,
        initial_value=numeric_val)

    assert simple_param() == numeric_val * scale + offset


def test_scaling_initial_value(simple_param, numeric_val):
    scale = 5
    offset = 3
    d = DelegateParameter(
        'test_delegate_parameter', simple_param, offset=offset, scale=scale)
    assert d() == (simple_param() - offset) / scale


def test_snapshot():
    p = Parameter('testparam', set_cmd=None, get_cmd=None,
                  offset=1, scale=2, initial_value=1)
    d = DelegateParameter('test_delegate_parameter', p, offset=3, scale=5,
                          initial_value=2)

    delegate_snapshot = d.snapshot()
    source_snapshot = delegate_snapshot.pop('source_parameter')
    assert source_snapshot == p.snapshot()
    assert delegate_snapshot['value'] == 2
    assert source_snapshot['value'] == 13


def test_set_source_cache_changes_delegate_cache(simple_param):
    """ Setting the cached value of the source parameter changes the
    delegate parameter cache accordingly.

    """
    offset = 4
    scale = 5
    d = DelegateParameter('d', simple_param, offset=offset, scale=scale)
    new_source_value = 3
    simple_param.cache.set(new_source_value)

    assert d.cache.get() == (new_source_value - offset) / scale


def test_set_source_cache_changes_delegate_get(simple_param):
    """ When the delegate parameter's ``get`` is called, the new
    value of the source propagates.

    """
    offset = 4
    scale = 5
    d = DelegateParameter('d', simple_param, offset=offset, scale=scale)
    new_source_value = 3

    simple_param.cache.set(new_source_value)

    assert d.get() == (new_source_value - offset) / scale


def test_set_delegate_cache_changes_source_cache(simple_param):
    offset = 4
    scale = 5
    d = DelegateParameter('d', simple_param, offset=offset, scale=scale)

    new_delegate_value = 2
    d.cache.set(new_delegate_value)

    assert simple_param.cache.get() == (new_delegate_value * scale + offset)


def test_set_delegate_cache_with_raw_value(simple_param):
    offset = 4
    scale = 5
    d = DelegateParameter('d', simple_param, offset=offset, scale=scale)

    new_delegate_value = 2
    d.cache._set_from_raw_value(new_delegate_value*scale + offset)

    assert simple_param.cache.get() == (new_delegate_value * scale + offset)
    assert d.cache.get(get_if_invalid=False) == new_delegate_value


def test_instrument_val_invariant_under_delegate_cache_set(
        make_observable_parameter, numeric_val):
    """
    Setting the cached value of the source parameter changes the delegate
    parameter. But it has no impact on the instrument value.
    """
    initial_value = numeric_val
    t = make_observable_parameter(
        'observable_parameter', initial_value=initial_value)
    new_source_value = 3
    t.cache.set(new_source_value)
    assert t.get_instr_val() == initial_value


def test_delegate_cache_pristine_if_not_set():
    p = Parameter('test')
    d = DelegateParameter('delegate', p)
    gotten_delegate_cache = d.cache.get(get_if_invalid=False)
    assert gotten_delegate_cache is None


def test_delegate_get_updates_cache(make_observable_parameter, numeric_val):
    initial_value = numeric_val
    t = make_observable_parameter(
        'observable_parameter', initial_value=initial_value)
    d = DelegateParameter('delegate', t)

    assert d() == initial_value
    assert d.cache.get() == initial_value
    assert t.get_instr_val() == initial_value


def test_delegate_parameter_get_and_snapshot_raises_with_none():
    """
    Test that a delegate parameter raises on get and snapshot if
    the source has a value of None and a scale is used.
    But works correctly if the source is remapped to a real parameter.

    """
    none_param = Parameter("None")
    source_param = Parameter('source', get_cmd=None, set_cmd=None, initial_value=2)
    delegate_param = DelegateParameter(name='delegate', source=none_param)
    delegate_param.scale = 2
    with pytest.raises(TypeError):
        delegate_param.get()
    with pytest.raises(TypeError):
        delegate_param.snapshot()
    assert delegate_param.cache._parameter.source.cache is none_param.cache
    delegate_param.source = source_param
    assert delegate_param.get() == 1
    assert delegate_param.snapshot()['value'] == 1
    assert delegate_param.cache._parameter.source.cache is source_param.cache


def test_raw_value_scaling(make_observable_parameter):
    """
    The :attr:`raw_value` will be deprecated soon,
    so other tests should not use it.
    """

    p = Parameter('testparam', set_cmd=None, get_cmd=None,
                  offset=1, scale=2)
    d = DelegateParameter('test_delegate_parameter', p, offset=3, scale=5)

    val = 1
    p(val)
    assert d() == (val - 3) / 5

    d(val)
    assert d.raw_value == val * 5 + 3
    assert d.raw_value == p()


def test_setting_initial_value_delegate_parameter():
    value = 10
    p = Parameter('testparam', set_cmd=None, get_cmd=None)
    d = DelegateParameter('test_delegate_parameter', p,
                          initial_value=value)
    assert p.cache.get(get_if_invalid=False) == value
    assert d.cache.get(get_if_invalid=False) == value


def test_setting_initial_cache_delegate_parameter():
    value = 10
    p = Parameter('testparam', set_cmd=None, get_cmd=None)
    d = DelegateParameter('test_delegate_parameter', p,
                          initial_cache_value=value)
    assert p.cache.get(get_if_invalid=False) == value
    assert d.cache.get(get_if_invalid=False) == value


def test_delegate_parameter_with_none_source_works_as_expected():
    delegate_param = DelegateParameter(name='delegate', source=None,
                                       scale=2, offset=1)
    _assert_none_source_is_correct(delegate_param)


@given(hst.floats(allow_nan=False, allow_infinity=False),
       hst.floats(allow_nan=False, allow_infinity=False).filter(lambda x: x != 0),
       hst.floats(allow_nan=False, allow_infinity=False))
def test_delegate_parameter_with_changed_source_snapshot_matches_value(value,
                                                                       scale,
                                                                       offset):
    delegate_param = DelegateParameter(name="delegate",
                                       source=None,
                                       scale=scale,
                                       offset=offset)
    source_parameter = Parameter(name="source",
                                 get_cmd=None,
                                 set_cmd=None,
                                 initial_value=value)
    _assert_none_source_is_correct(delegate_param)
    delegate_param.source = source_parameter
    calc_value = (value - offset) / scale
    assert delegate_param.cache.get(get_if_invalid=False) == calc_value
    assert delegate_param.source.cache.get(get_if_invalid=False) == value
    snapshot = delegate_param.snapshot()
    # disregard timestamp that might be slightly different
    snapshot["source_parameter"].pop("ts")
    source_snapshot = source_parameter.snapshot()
    source_snapshot.pop("ts")
    assert snapshot["source_parameter"] == source_snapshot
    assert snapshot["value"] == calc_value
    assert delegate_param.get() == calc_value
    # now remove the source again
    delegate_param.source = None
    _assert_none_source_is_correct(delegate_param)
    _assert_delegate_cache_none_source(delegate_param)


def _assert_none_source_is_correct(delegate_param):
    with pytest.raises(TypeError):
        delegate_param.get()
    with pytest.raises(TypeError):
        delegate_param.set(1)
    snapshot = delegate_param.snapshot()
    assert snapshot["source_parameter"] is None
    assert "value" not in snapshot.keys()
    snapshot.pop("ts")
    updated_snapshot = delegate_param.snapshot(update=True)
    updated_snapshot.pop("ts")
    assert snapshot == updated_snapshot


def _assert_delegate_cache_none_source(delegate_param):
    with pytest.raises(TypeError):
        delegate_param.cache.set(1)
    with pytest.raises(TypeError):
        delegate_param.cache.get()
    with pytest.raises(TypeError):
        delegate_param.cache.raw_value
    assert delegate_param.cache.max_val_age is None
    assert delegate_param.cache.timestamp is None


@pytest.mark.parametrize("snapshot_value", [True, False])
@pytest.mark.parametrize("gettable,get_cmd", [(True, None), (False, False)])
@pytest.mark.parametrize("settable,set_cmd", [(True, None), (False, False)])
def test_gettable_settable_snapshotget_delegate_parameter(gettable, get_cmd,
                                                          settable, set_cmd,
                                                          snapshot_value):
    """
    Test that gettable, settable and snapshot_get are correctly reflected
    in the DelegateParameter
    """
    source_param = Parameter("source", get_cmd=get_cmd,
                             set_cmd=set_cmd, snapshot_value=snapshot_value)
    delegate_param = DelegateParameter("delegate", source=source_param)
    assert delegate_param.gettable is gettable
    assert delegate_param.settable is settable
    assert delegate_param._snapshot_value is snapshot_value


@pytest.mark.parametrize("snapshot_value", [True, False])
@pytest.mark.parametrize("gettable,get_cmd", [(True, None), (False, False)])
@pytest.mark.parametrize("settable,set_cmd", [(True, None), (False, False)])
def test_gettable_settable_snapshotget_delegate_parameter_2(gettable, get_cmd,
                                                            settable, set_cmd,
                                                            snapshot_value):
    """
    Test that gettable/settable and snapshot_get are updated correctly
    when source changes
    """
    source_param = Parameter("source", get_cmd=get_cmd, set_cmd=set_cmd,
                             snapshot_value=snapshot_value)
    delegate_param = DelegateParameter("delegate", source=None)
    delegate_param.source = source_param
    assert delegate_param.gettable is gettable
    assert delegate_param.settable is settable
    assert delegate_param._snapshot_value is snapshot_value


def test_initial_value_and_none_source_raises():
    with pytest.raises(KeyError, match="It is not allowed to supply"
                                       " 'initial_value' or"
                                       " 'initial_cache_value'"):
        DelegateParameter("delegate", source=None, initial_value=1)
    with pytest.raises(KeyError, match="It is not allowed to supply"
                                       " 'initial_value' or "
                                       "'initial_cache_value'"):
        DelegateParameter("delegate", source=None, initial_cache_value=1)


def test_delegate_parameter_change_source_reflected_in_label_and_unit():
    delegate_param = DelegateParameter("delegate", source=None)
    source_param_1 = Parameter("source1", label="source 1", unit="unit1")
    source_param_2 = Parameter("source2", label="source 2", unit="unit2")

    assert delegate_param.label == "delegate"
    assert delegate_param.unit == ""
    delegate_param.source = source_param_1
    assert delegate_param.label == "source 1"
    assert delegate_param.unit == "unit1"
    delegate_param.source = source_param_2
    assert delegate_param.label == "source 2"
    assert delegate_param.unit == "unit2"
    delegate_param.source = None
    assert delegate_param.label == "delegate"
    assert delegate_param.unit == ""


def test_delegate_parameter_fixed_label_unit_unchanged():
    delegate_param = DelegateParameter("delegate",
                                       label="delegatelabel",
                                       unit="delegateunit",
                                       source=None)
    source_param_1 = Parameter("source1", label="source 1", unit="unit1")
    source_param_2 = Parameter("source2", label="source 2", unit="unit2")

    assert delegate_param.label == "delegatelabel"
    assert delegate_param.unit == "delegateunit"
    delegate_param.source = source_param_1
    assert delegate_param.label == "delegatelabel"
    assert delegate_param.unit == "delegateunit"
    delegate_param.source = source_param_2
    assert delegate_param.label == "delegatelabel"
    assert delegate_param.unit == "delegateunit"
    delegate_param.source = None
    assert delegate_param.label == "delegatelabel"
    assert delegate_param.unit == "delegateunit"


def test_cache_invalidation():
    value = 10
    p = BetterGettableParam('testparam', set_cmd=None, get_cmd=None)
    d = DelegateParameter('test_delegate_parameter', p,
                          initial_cache_value=value)
    assert p._get_count == 0
    assert d.cache.get() == value
    assert p._get_count == 0

    assert d.cache.valid is True
    assert p.cache.valid is True

    d.cache.invalidate()

    assert d.cache.valid is False
    assert p.cache.valid is False

    d.cache.get()
    assert p._get_count == 1

    assert d.cache.valid is True
    assert p.cache.valid is True


def test_cache_no_source():
    d = DelegateParameter('test_delegate_parameter', source=None)

    assert d.cache.valid is False
    assert d.cache.timestamp is None
    assert d.cache.max_val_age is None

    with pytest.raises(
            TypeError,
            match="Cannot get the cache of a "
                  "DelegateParameter that delegates to None"):
        d.cache.get()

    d.cache.invalidate()


def test_underlying_instrument_property_for_delegate_parameter():
    p = BetterGettableParam('testparam', set_cmd=None, get_cmd=None)
    d = DelegateParameter('delegate_parameter_with_source', p)

    assert d.underlying_instrument is p.root_instrument

    d = DelegateParameter('delegate_parameter_without_source', source=None)
    assert d.underlying_instrument is None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, Generic, Mapping, Optional, Tuple, TypeVar

import numpy as np

from qcodes.dataset.descriptions.rundescriber import RunDescriber
from qcodes.dataset.sqlite.connection import ConnectionPlus
from qcodes.dataset.sqlite.queries import completed, load_new_data_for_rundescriber
from qcodes.utils.deprecate import deprecate

from .exporters.export_to_pandas import (
    load_to_concatenated_dataframe,
    load_to_dataframe_dict,
)
from .exporters.export_to_xarray import (
    load_to_xarray_dataarray_dict,
    load_to_xarray_dataset,
)

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr

    from .data_set import DataSet, ParameterData
    from .data_set_protocol import DataSetProtocol

DatasetType = TypeVar("DatasetType", bound="DataSetProtocol", covariant=True)

log = logging.getLogger(__name__)


class DataSetCache(Generic[DatasetType]):
    """
    The DataSetCache contains a in memory representation of the
    data in this dataset as well a a method to progressively read data
    from the db as it is written and methods to append data as it is received
    without writing it to disk. The cache can either be loaded from the db
    or produced as an in memory cache. It is not possible to combine these
    two ways of producing a dataset cache. The cache is available in the
    same formats as :py:class:`.DataSet.get_parameter_data` and
    :py:class:`.DataSet.to_pandas_dataframe_dict`
    """

    def __init__(self, dataset: DatasetType):
        self._dataset = dataset
        self._data: ParameterData = {}
        #: number of rows read per parameter tree (by the name of the dependent parameter)
        self._read_status: Dict[str, int] = {}
        #: number of rows written per parameter tree (by the name of the dependent parameter)
        self._write_status: Dict[str, Optional[int]] = {}
        self._loaded_from_completed_ds = False
        self._live: Optional[bool] = None

    @property
    def rundescriber(self) -> RunDescriber:
        return self._dataset.description

    @property
    def live(self) -> Optional[bool]:
        """
        If true this cache has been produced by appending data as measured.
        If false the data has been read from disk.
        If None, then the cache does not yet have any data.
        """
        return self._live

    def data(self) -> ParameterData:
        """
        Loads data from the database on disk if needed and returns
        the cached data. The cached data is in almost the same format as
        :py:class:`.DataSet.get_parameter_data`. However if a shape is provided
        as part of the dataset metadata and fewer datapoints than expected are
        returned the missing values will be replaced by `NaN` or zeroes
        depending on the datatype.

        Returns:
            The cached dataset.
        """
        if not self.live:
            self.load_data_from_db()

        return self._data

    def load_data_from_db(self) -> None:
        """
        Load the data from an on-disk format in case the cache is not live

        Should be implemented in a specific subclass that knows how to read data
        from disk
        """

    def add_data(self, new_data: Mapping[str, Mapping[str, np.ndarray]]) -> None:
        if self.live is False:
            raise RuntimeError(
                "Cannot append live data to a dataset that has "
                "been fully or partially loaded from a database."
            )

        expanded_data = {}
        for param_name, single_param_dict in new_data.items():
            expanded_data[param_name] = _expand_single_param_dict(
                single_param_dict
            )

        (
            self._write_status,
            self._data
        ) = append_shaped_parameter_data_to_existing_arrays(
            self.rundescriber,
            self._write_status,
            self._data,
            new_data=expanded_data
        )

        if not all(status is None for status in self._write_status.values()):
            self._live = True

    def to_pandas_dataframe_dict(self) -> Dict[str, pd.DataFrame]:
        """
        Convert the cached dataset to Pandas dataframes. The returned dataframes
        are in the same format :py:class:`.DataSet.to_pandas_dataframe_dict`.

        Returns:
            A dict from parameter name to Pandas Dataframes. Each dataframe
            represents one parameter tree.
        """
        data = self.data()
        return load_to_dataframe_dict(data)

    def to_pandas_dataframe(self) -> pd.DataFrame:
        """
        Convert the cached dataset to Pandas dataframes. The returned dataframes
        are in the same format :py:class:`.DataSet.to_pandas_dataframe_dict`.

        Returns:
            A dict from parameter name to Pandas Dataframes. Each dataframe
            represents one parameter tree.
        """
        data = self.data()
        return load_to_concatenated_dataframe(data)

    @deprecate(alternative="to_pandas_dataframe or to_pandas_dataframe_dict")
    def to_pandas(self) -> Dict[str, pd.DataFrame]:
        """
        Returns the values stored in the :class:`.dataset.data_set.DataSet` as a
        concatenated :py:class:`pandas.DataFrame` s

        The DataFrame contains a column for the data and is indexed by a
        :py:class:`pandas.MultiIndex` formed from all the setpoints
        of the parameter.

        Note that if the dataset contains data for multiple parameters that do
        not share the same setpoints it is recommended to use
        :py:class:`.to_pandas_dataframe_dict`

        Returns:
            :py:class:`pandas.DataFrame` s with the requested parameter as
            a column and a indexed by a :py:class:`pandas.MultiIndex` formed
            by the dependencies.
        """
        return self.to_pandas_dataframe_dict()

    def to_xarray_dataarray_dict(self) -> Dict[str, xr.DataArray]:
        """
        Returns the values stored in the :class:`.dataset.data_set.DataSet` as a dict of
        :py:class:`xr.DataArray` s
        Each element in the dict is indexed by the names of the dependent parameters.

        Returns:
            Dictionary from requested parameter names to :py:class:`xr.DataArray` s
            with the requested parameter(s) as a column(s) and coordinates
            formed by the dependencies.

        """
        data = self.data()
        return load_to_xarray_dataarray_dict(self._dataset, data)

    def to_xarray_dataset(self) -> xr.Dataset:
        """
        Returns the values stored in the :class:`.dataset.data_set.DataSet` as a
        :py:class:`xr.Dataset` object.

        Note that if the dataset contains data for multiple parameters that do
        not share the same setpoints it is recommended to use
        :py:class:`.to_xarray_dataarray_dict`

        Returns:
            :py:class:`xr.Dataset` with the requested parameter(s) data as
            :py:class:`xr.DataArray` s and coordinates formed by the dependencies.

        """
        data = self.data()
        return load_to_xarray_dataset(self._dataset, data)


def load_new_data_from_db_and_append(
    conn: ConnectionPlus,
    table_name: str,
    rundescriber: RunDescriber,
    write_status: Mapping[str, Optional[int]],
    read_status: Mapping[str, int],
    existing_data: Mapping[str, Mapping[str, np.ndarray]],
) -> Tuple[Dict[str, Optional[int]], Dict[str, int], Dict[str, Dict[str, np.ndarray]]]:
    """
    Append any new data in the db to an already existing datadict and return the merged
    data.

    Args:
        conn: The connection to the sqlite database
        table_name: The name of the table the data is stored in
        rundescriber: The rundescriber that describes the run
        write_status: Mapping from dependent parameter name to number of rows
          written to the cache previously.
        read_status: Mapping from dependent parameter name to number of rows
          read from the db previously.
        existing_data: Mapping from dependent parameter name to mapping
          from parameter name to numpy arrays that the data should be
          inserted into.
          appended to.

    Returns:
        Updated write and read status, and the updated ``data``

    """
    new_data, updated_read_status = load_new_data_for_rundescriber(
        conn, table_name, rundescriber, read_status
    )

    (updated_write_status,
     merged_data) = append_shaped_parameter_data_to_existing_arrays(
        rundescriber,
        write_status,
        existing_data,
        new_data
    )
    return updated_write_status, updated_read_status, merged_data


def append_shaped_parameter_data_to_existing_arrays(
    rundescriber: RunDescriber,
    write_status: Mapping[str, Optional[int]],
    existing_data: Mapping[str, Mapping[str, np.ndarray]],
    new_data: Mapping[str, Mapping[str, np.ndarray]],
) -> Tuple[Dict[str, Optional[int]], Dict[str, Dict[str, np.ndarray]]]:
    """
    Append datadict to an already existing datadict and return the merged
    data.

    Args:
        rundescriber: The rundescriber that describes the run
        write_status: Mapping from dependent parameter name to number of rows
          written to the cache previously.
        new_data: Mapping from dependent parameter name to mapping
          from parameter name to numpy arrays that the data should be
          appended to.
        existing_data: Mapping from dependent parameter name to mapping
          from parameter name to numpy arrays of new data.

    Returns:
        Updated write and read status, and the updated ``data``
    """
    parameters = tuple(ps.name for ps in
                       rundescriber.interdeps.non_dependencies)
    merged_data = {}

    updated_write_status = dict(write_status)

    for meas_parameter in parameters:

        existing_data_1_tree = existing_data.get(meas_parameter, {})

        new_data_1_tree = new_data.get(meas_parameter, {})

        shapes = rundescriber.shapes
        if shapes is not None:
            shape = shapes.get(meas_parameter, None)
        else:
            shape = None

        (merged_data[meas_parameter],
         updated_write_status[meas_parameter]) = _merge_data(
            existing_data_1_tree,
            new_data_1_tree,
            shape,
            single_tree_write_status=write_status.get(meas_parameter),
            meas_parameter=meas_parameter
        )
    return updated_write_status, merged_data


def _merge_data(existing_data: Mapping[str, np.ndarray],
                new_data: Mapping[str, np.ndarray],
                shape: Optional[Tuple[int, ...]],
                single_tree_write_status: Optional[int],
                meas_parameter: str,
                ) -> Tuple[Dict[str, np.ndarray], Optional[int]]:

    subtree_merged_data = {}
    subtree_parameters = set(existing_data.keys()) | set(new_data.keys())
    new_write_status: Optional[int]
    single_param_merged_data, new_write_status = _merge_data_single_param(
        existing_data.get(meas_parameter),
        new_data.get(meas_parameter),
        shape,
        single_tree_write_status
    )
    if single_param_merged_data is not None:
        subtree_merged_data[meas_parameter] = single_param_merged_data

    for subtree_param in subtree_parameters:
        if subtree_param != meas_parameter:

            single_param_merged_data, new_write_status = _merge_data_single_param(
                existing_data.get(subtree_param),
                new_data.get(subtree_param),
                shape,
                single_tree_write_status
            )
            if single_param_merged_data is not None:
                subtree_merged_data[subtree_param] = single_param_merged_data

    return subtree_merged_data, new_write_status


def _merge_data_single_param(
        existing_values: Optional[np.ndarray],
        new_values: Optional[np.ndarray],
        shape: Optional[Tuple[int, ...]],
        single_tree_write_status: Optional[int]) -> Tuple[Optional[np.ndarray], Optional[int]]:
    merged_data: Optional[np.ndarray]
    if existing_values is not None and new_values is not None:
        (merged_data,
         new_write_status) = _insert_into_data_dict(
            existing_values,
            new_values,
            single_tree_write_status,
            shape=shape
        )
    elif new_values is not None:
        (merged_data,
         new_write_status) = _create_new_data_dict(
            new_values,
            shape
        )
    elif existing_values is not None:
        merged_data = existing_values
        new_write_status = single_tree_write_status
    else:
        merged_data = None
        new_write_status = None
    return merged_data, new_write_status


def _create_new_data_dict(new_values: np.ndarray,
                          shape: Optional[Tuple[int, ...]]
                          ) -> Tuple[np.ndarray, int]:
    if shape is None:
        return new_values, new_values.size
    else:
        n_values = new_values.size
        data = np.zeros(shape, dtype=new_values.dtype)

        if new_values.dtype.kind == "f":
            data[:] = np.nan
        elif new_values.dtype.kind == "c":
            data[:] = np.nan + 1j * np.nan

        data.ravel()[0:n_values] = new_values.ravel()
        return data, n_values


def _insert_into_data_dict(
        existing_values: np.ndarray,
        new_values: np.ndarray,
        write_status: Optional[int],
        shape: Optional[Tuple[int, ...]]
) -> Tuple[np.ndarray, Optional[int]]:
    if shape is None or write_status is None:
        try:
            data = np.append(existing_values, new_values, axis=0)
        except ValueError:
            # we cannot append into a ragged array so make that manually
            n_existing = existing_values.shape[0]
            n_new = new_values.shape[0]
            n_rows = n_existing + n_new
            data = np.ndarray((n_rows,), dtype=object)

            for i in range(n_existing):
                data[i] = np.atleast_1d(existing_values[i])
            for i, j in enumerate(range(n_existing, n_existing+n_new)):
                data[j] = np.atleast_1d(new_values[i])
        return data, None
    else:
        if existing_values.dtype.kind in ('U', 'S'):
            # string type arrays may be too small for the new data
            # read so rescale if needed.
            if new_values.dtype.itemsize > existing_values.dtype.itemsize:
                existing_values = existing_values.astype(new_values.dtype)
        n_values = new_values.size
        new_write_status = write_status+n_values
        if new_write_status > existing_values.size:
            log.warning(f"Incorrect shape of dataset: Dataset is expected to "
                        f"contain {existing_values.size} points but trying to "
                        f"add an amount of data that makes it contain {new_write_status} points. Cache will "
                        f"be flattened into a 1D array")
            return (np.append(existing_values.flatten(),
                              new_values.flatten(), axis=0),
                    new_write_status)
        else:
            existing_values.ravel()[write_status:new_write_status] = new_values.ravel()
            return existing_values, new_write_status


def _expand_single_param_dict(
        single_param_dict: Mapping[str, np.ndarray]
) -> Dict[str, np.ndarray]:
    sizes = {name: array.size for name, array in single_param_dict.items()}
    maxsize = max(sizes.values())
    max_names = tuple(name for name, size in sizes.items() if size == maxsize)

    expanded_param_dict = {}
    for name, array in single_param_dict.items():
        if name in max_names:
            expanded_param_dict[name] = array
        else:
            assert array.size == 1
            expanded_param_dict[name] = np.full_like(
                single_param_dict[max_names[0]],
                array.ravel()[0], dtype=array.dtype
            )

    return expanded_param_dict


class DataSetCacheWithDBBackend(DataSetCache["DataSet"]):
    def load_data_from_db(self) -> None:
        """
        Loads data from the dataset into the cache.
        If new data has been added to the dataset since the last time
        this method was called, calling this method again would load
        that new portion of the data and append to the already loaded data.
        If the dataset is marked completed and data has already been loaded
        no load will be performed.
        """
        if self.live:
            raise RuntimeError(
                "Cannot load data into this cache from the "
                "database because this dataset is being built "
                "in-memory."
            )

        if self._loaded_from_completed_ds:
            return
        self._dataset.completed = completed(self._dataset.conn, self._dataset.run_id)
        if self._dataset.completed:
            self._loaded_from_completed_ds = True

        (
            self._write_status,
            self._read_status,
            self._data,
        ) = load_new_data_from_db_and_append(
            self._dataset.conn,
            self._dataset.table_name,
            self.rundescriber,
            self._write_status,
            self._read_status,
            self._data,
        )
        if not all(status is None for status in self._write_status.values()):
            self._live = False



from typing import Dict, Optional

from qcodes import IPInstrument
from qcodes.instrument.channel import ChannelList, InstrumentChannel
from qcodes.utils import validators as vals


class MC_channel(InstrumentChannel):
    def __init__(self, parent: "RC_SPDT", name: str, channel_letter: str):
        """
        Args:
            parent: The instrument the channel is a part of
            name: the name of the channel
            channel_letter: channel letter ['a', 'b', 'c' or 'd'])
        """

        super().__init__(parent, name)
        self.channel_letter = channel_letter.upper()
        _chanlist = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
        self.channel_number = _chanlist.index(channel_letter)

        self.add_parameter('switch',
                           label='switch',
                           set_cmd=self._set_switch,
                           get_cmd=self._get_switch,
                           vals=vals.Ints(1, 2)
                           )

    def _set_switch(self, switch: int) -> None:
        self.write(f"SET{self.channel_letter}={switch-1}")

    def _get_switch(self) -> int:
        val = int(self.ask('SWPORT?'))
        # select out bit in return number
        # corisponding to channel switch configuration
        # LSB corrisponds to Chan A etc
        ret = (val >> self.channel_number) & 1
        return ret+1


class RC_SPDT(IPInstrument):
    """
    Mini-Circuits SPDT RF switch

    Args:
        name: the name of the instrument
        address: ip address ie "10.0.0.1"
        port: port to connect to default Telnet:23
    """
    def __init__(self, name: str, address: str, port: int = 23):
        super().__init__(name, address, port)
        self.flush_connection()

        channels = ChannelList(self, "Channels", MC_channel,
                               snapshotable=False)

        _chanlist = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
        _max_channel_number = int(self.IDN()['model'][3])
        _chanlist = _chanlist[0:_max_channel_number]

        for c in _chanlist:
            channel = MC_channel(self, f'channel_{c}', c)
            channels.append(channel)
            self.add_submodule(f'channel_{c}', channel)
        channels.lock()
        self.add_submodule('channels', channels)

        self.connect_message()

    def ask(self, cmd: str) -> str:
        ret = self.ask_raw(cmd)
        ret = ret.strip()
        return ret

    def get_idn(self) -> Dict[str, Optional[str]]:

        fw = self.ask('FIRMWARE?')
        MN = self.ask('MN?')
        SN = self.ask('SN?')

        id_dict: Dict[str, Optional[str]] = {
            'firmware': fw,
            'model': MN[3:],
            'serial': SN[3:],
            'vendor': 'Mini-Circuits'
        }
        return id_dict



"""
This module contains driver and related infrastructure for Keysight B1500
Semiconductor Parameter Analyzer.

The structure of the module is as follows:

  - The driver of the Parameter Analyzer is located in ``KeysightB1500_base.py``
    file.
  - ``KeysightB1500_module.py`` file implements a class that is common for
    all the modules that the Parameter Analyzer may support.
  - Other ``Keysight*.py`` files implement classes for specific modules that
    the Parameter Analyzer may support.
  - ``message_builder.py`` and ``constants.py`` implement
    infrastructure for low-level interfacing with the Parameter Analyzer.

"""
from .message_builder import MessageBuilder
from .KeysightB1500_base import KeysightB1500
from . import constants

__all__ = ['KeysightB1500', 'MessageBuilder', 'constants']



import warnings
from typing import TYPE_CHECKING, Any, List

import numpy

from qcodes.instrument.parameter import ParameterWithSetpoints
from .message_builder import MessageBuilder
from . import constants
from .KeysightB1500_module import fmt_response_base_parser, _FMTResponse, \
    MeasurementNotTaken, convert_dummy_val_to_nan

if TYPE_CHECKING:
    from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1517A \
        import B1517A
    from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1500_base \
        import KeysightB1500


class SamplingMeasurement(ParameterWithSetpoints):
    """
    Performs sampling measurement using semiconductor
    parameter analyzer B1500A.
    """

    _timeout_response_factor = 10.00
    # This factor is a bit higher than the ratio between
    # the measured measurement-time and the calculated measurement
    # (from the user input). Check :get_raw: method to find its usage.

    def __init__(self, name: str, **kwargs: Any):
        super().__init__(name, **kwargs)

        self.instrument: "B1517A"
        self.root_instrument: "KeysightB1500"

        self.data = _FMTResponse(None, None, None, None)

    def get_raw(self) -> numpy.ndarray:
        """
        This performs sampling  measurements. However since the measurement
        time can vary from few seconds to hundreds of minutes we first set
        the visa time out.
        The visa time-out should be longer than the time it takes to finish
        the sampling measurement. The reason is that while measurement is
        running the data keeps on appending in the buffer of SPA. Only when
        the measurement is finished the data is returned to the VISA handle.
        Hence during this time the VISA is idle and waiting for the response.
        If the timeout is lower than the total run time of the measurement,
        VISA will give error.
        We set the Visa timeout to be the `measurement_time` times the
        `_timeout_response_factor`. Strictly speaking the timeout should be
        just longer than the measurement time.
        """

        measurement_time = self.instrument._total_measurement_time()
        time_out = measurement_time * self._timeout_response_factor
        default_timeout = self.root_instrument.timeout()

        # if time out to be set is lower than the default value
        # then keep default
        if time_out < default_timeout:
            time_out = default_timeout

        self.root_instrument.write(MessageBuilder().fmt(1, 0).message)

        with self.root_instrument.timeout.set_to(time_out):
            raw_data = self.root_instrument.ask(
                MessageBuilder().xe().message)

        self.data = fmt_response_base_parser(raw_data)
        convert_dummy_val_to_nan(self.data)
        return numpy.array(self.data.value)

    def compliance(self) -> List[int]:
        """
        check for the status other than "N" (normal) and output the
        number of data values which were not measured under "N" (normal)
        status.

        For the list of all the status values and their meaning refer to
        :class:`.constants.MeasurementStatus`.

        """

        if self.data.status is None:
            raise MeasurementNotTaken('First run sampling_measurement'
                                      ' method to generate the data')
        else:
            data = self.data
            total_count = len(data.status)
            normal_count = data.status.count(constants.MeasurementStatus.N.name)
            exception_count = total_count - normal_count
            if total_count == normal_count:
                print('All measurements are normal')
            else:
                indices = [i for i, x in enumerate(data.status)
                           if x == "C" or x == "T"]
                warnings.warn(f'{str(exception_count)} measurements were '
                              f'out of compliance at {str(indices)}')

            compliance_list = [constants.MeasurementError[key].value
                               for key in data.status]
            return compliance_list



"""
The dataset module contains code related to storage and retrieval of data to
and from disk
"""
from .measurements import Measurement
from .data_set import new_data_set, load_by_counter, load_by_id,  \
    load_by_run_spec, load_by_guid
from .experiment_container import new_experiment, load_experiment,  \
    load_experiment_by_name, load_last_experiment, experiments,  \
    load_or_create_experiment
from .sqlite.settings import SQLiteSettings
from .descriptions.param_spec import ParamSpec
from .sqlite.database import initialise_database

# flake8: noqa (we don't need the "<...> imported but unused" error)

from .sqlite.database import initialise_or_create_database_at



from typing import Any
from .private.Keysight_344xxA_submodules import _Keysight_344xxA


class Keysight_34410A(_Keysight_344xxA):
    """
    This is the qcodes driver for the Keysight 34410A Multimeter
    """
    def __init__(self, name: str, address: str, silent: bool = False,
                 **kwargs: Any):
        super().__init__(name, address, silent, **kwargs)



import math

import numpy as np
import pytest
from qcodes.utils.validators import Multiples

from .conftest import AClass, a_func

divisors = [3, 7, 11, 13]
not_divisors = [0, -1, -5, -1e15, 0.1, -0.1, 1.0, 3.5,
                -2.3e6, 5.5e15, 1.34e-10, -2.5e-5,
                math.pi, math.e, '', None, float("nan"), float("inf"),
                -float("inf"), '1', [], {}, [1, 2], {1: 1}, b'good',
                AClass, AClass(), a_func]
multiples = [0, 1, 10, -1, 100, 1000000, int(-1e15), int(1e15),
             # warning: True==1 and False==0 - we *could* prohibit these, using
             # isinstance(v, bool)
             True, False,
             # numpy scalars
             np.int64(2)]
not_multiples = [0.1, -0.1, 1.0, 3.5, -2.3e6, 5.5e15, 1.34e-10, -2.5e-5,
                 math.pi, math.e, '', None, float("nan"), float("inf"),
                 -float("inf"), '1', [], {}, [1, 2], {1: 1}, b'good',
                 AClass, AClass(), a_func]


def test_divisors():
    for d in divisors:
        n = Multiples(divisor=d)
        for v in [d * e for e in multiples]:
            n.validate(v)

        for v in multiples:
            if v == 0:
                continue
            with pytest.raises(ValueError):
                n.validate(v)

        for v in not_multiples:
            with pytest.raises(TypeError):
                n.validate(v)

    for d in not_divisors:
        with pytest.raises(TypeError):
            n = Multiples(divisor=d)

    n = Multiples(divisor=3, min_value=1, max_value=10)
    assert repr(n) == '<Ints 1<=v<=10, Multiples of 3>'


def test_valid_values():
    for d in divisors:
        n = Multiples(divisor=d)
        for num in n.valid_values:
            n.validate(num)



import logging
import re
from typing import Callable, List, Optional, Tuple, Union

from qcodes import InstrumentChannel, VisaInstrument, validators

from .keysight_34980a_submodules import KeysightSwitchMatrixSubModule


class Keysight34934A(KeysightSwitchMatrixSubModule):
    """
    Create an instance for module 34933A.
    Args:
        parent: the system which the module is installed on
        name: user defined name for the module
        slot: the slot the module is installed
    """
    def __init__(
            self,
            parent: Union[VisaInstrument, InstrumentChannel],
            name: str,
            slot: int
    ) -> None:

        super().__init__(parent, name, slot)

        self.add_parameter(name='protection_mode',
                           get_cmd=self._get_relay_protection_mode,
                           set_cmd=self._set_relay_protection_mode,
                           vals=validators.Enum('AUTO100',
                                                 'AUTO0',
                                                 'FIX',
                                                 'ISO'),
                           docstring='get and set the relay protection mode.'
                                     'The fastest switching speeds for relays'
                                     'in a given signal path are achieved using'
                                     'the FIXed or ISOlated modes, followed'
                                     'by the AUTO100 and AUTO0 modes.'
                                     'There may be a maximum of 200 Ohm of'
                                     'resistance, which can only be bypassed'
                                     'by "AUTO0" mode. See manual and'
                                     'programmer''s reference for detail.')

        layout = self.ask(f'SYSTEM:MODule:TERMinal:TYPE? {self.slot}')
        self._is_locked = (layout == 'NONE')
        if self._is_locked:
            logging.warning(
                f"For slot {slot}, no configuration module"
                f"connected, or safety interlock jumper removed. "
                "Making any connection is not allowed"
            )
            config = self.ask(f"SYST:CTYP? {slot}").strip('"').split(",")[1]
            layout = config.split("-")[1]
        self.row, self.column = (int(num) for num in re.findall(r"\d+", layout))

    def write(self, cmd: str) -> None:
        """
        When the module is safety interlocked, users can not make any
        connections. There will be no effect when try to connect any channels.
        """
        if self._is_locked:
            logging.warning("Warning: no configuration module connected, "
                            "or safety interlock enabled. "
                            "Making any connection is not allowed")
        return self.parent.write(cmd)

    def validate_value(self, row: int, column: int) -> None:
        """
        to check if the row and column number is within the range of the
        module layout.

        Args:
            row: row value
            column: column value
        """
        if (row > self.row) or (column > self.column):
            raise ValueError('row/column value out of range')

    def _get_relay_protection_mode(self) -> str:
        return self.ask(f'SYSTem:MODule:ROW:PROTection? {self.slot}')

    def _set_relay_protection_mode(self, mode: str) -> None:
        self.write(f'SYSTem:MODule:ROW:PROTection {self.slot}, {mode}')

    def to_channel_list(
            self,
            paths: List[Tuple[int, int]],
            wiring_config: Optional[str] = ''
    ) -> str:
        """
        convert the (row, column) pair to a 4-digit channel number 'sxxx', where
        s is the slot number, xxx is generated from the numbering function.

        Args:
            paths: list of channels to connect [(r1, c1), (r2, c2), (r3, c3)]
            wiring_config: for 1-wire matrices, values are 'MH', 'ML';
                                 for 2-wire matrices, values are 'M1H', 'M2H',
                                 'M1L', 'M2L'

        Returns:
            in the format of '(@sxxx, sxxx, sxxx, sxxx)', where sxxx is a
            4-digit channel number
        """
        numbering_function = self.get_numbering_function(
            self.row,
            self.column,
            wiring_config
        )

        channels = []
        for row, column in paths:
            channel = f'{self.slot}{numbering_function(row, column)}'
            channels.append(channel)
        channel_list = f"(@{','.join(channels)})"
        return channel_list

    @staticmethod
    def get_numbering_function(
            rows: int,
            columns: int,
            wiring_config: Optional[str] = ''
    ) -> Callable[[int, int], str]:
        """
        to select the correct numbering function based on the matrix layout.
        On P168 of the user's guide for Agilent 34934A High Density Matrix
        Module:
        http://literature.cdn.keysight.com/litweb/pdf/34980-90034.pdf
        there are eleven equations. This function here simplifies them to one.

        Args:
            rows: the total row number of the matrix module
            columns: the total column number of the matrix module
            wiring_config: wiring configuration for 1 or 2 wired matrices

        Returns:
            The numbering function to convert row and column in to a 3-digit
            number
        """
        layout = f'{rows}x{columns}'
        available_layouts = {
            "4x32": ["M1H", "M2H", "M1L", "M2L"],
            "4x64": ["MH", "ML"],
            "4x128": [''],
            "8x32": ["MH", "ML"],
            "8x64": [''],
            "16x32": ['']
        }

        if layout not in available_layouts:
            raise ValueError(f"Unsupported layout: {layout}")

        if wiring_config not in available_layouts[layout]:
            raise ValueError(
                f"Invalid wiring config '{wiring_config}' for layout {layout}"
            )

        offsets = {
            "M1H": 0,
            "M2H": 1,
            "M1L": 2,
            "M2L": 3,
            "MH": 0,
            "ML": 1
        }

        offset = 0
        if wiring_config != '':
            offset = offsets[wiring_config] * columns

        channels_per_row = 800 / rows
        offset += 100 - int(channels_per_row)

        def numbering_function(row: int, col: int) -> str:
            return str(int(channels_per_row * row + col + offset))

        return numbering_function



from typing import Dict, Optional

from qcodes import IPInstrument
from qcodes.utils import validators as vals
from qcodes.instrument.channel import InstrumentChannel, ChannelList
import math


class MC_channel(InstrumentChannel):
    def __init__(self, parent: IPInstrument, name: str, channel_letter: str):
        """
        Args:
            parent: The instrument the channel is a part of
            name: the name of the channel
            channel_letter: channel letter ['a', 'b'])
        """

        super().__init__(parent, name)
        self.channel_letter = channel_letter.upper()
        chanlist = ['a', 'b']
        self.channel_number = chanlist.index(channel_letter)

        self.add_parameter('switch',
                           label='switch',
                           set_cmd=self._set_switch,
                           get_cmd=self._get_switch,
                           vals=vals.Ints(0, 4)
                           )

    def _set_switch(self, switch: int) -> None:
        if len(self._parent.channels) > 1:
            current_switchstate = int(self.ask('SWPORT?'))
            mask = 0xF << (4 * (1 - self.channel_number))
            current_switchstate = current_switchstate & mask
        else:
            current_switchstate = 0
        # getting the current switch state of
        # the other channel if more than one channel used

        if switch == 0:
            val = 0
        else:
            val = 1 << (switch-1)
        val = val << (4 * self.channel_number)
        # only one bit in each nibble can be set
        # corrisponding to each switch state
        # setting more than one will be ignored
        val = val | current_switchstate
        self.write(f'SETP={val}')
        # the 'SP4T' command wont work on early devices (FW < C8)

    def _get_switch(self) -> int:
        val = int(self.ask('SWPORT?'))
        val = (val >> (4*self.channel_number)) & 0xF
        if val == 0:
            return 0
        ret = int(math.log2(val) + 1)
        return ret


class RC_SP4T(IPInstrument):
    """
    Mini-Circuits SP4T RF switch

    Args:
        name: the name of the instrument
        address: ip address ie "10.0.0.1"
        port: port to connect to default Telnet:23
    """
    def __init__(self, name: str, address: str, port: int = 23):
        super().__init__(name, address, port)
        self.flush_connection()

        channels = ChannelList(self, "Channels", MC_channel,
                               snapshotable=False)

        _chanlist = ['a', 'b']
        _max_channel_number = int(self.IDN()['model'][3])
        _chanlist = _chanlist[0:_max_channel_number]

        for c in _chanlist:
            channel = MC_channel(self, f'channel_{c}', c)
            channels.append(channel)
            self.add_submodule(f'channel_{c}', channel)
        channels.lock()
        self.add_submodule('channels', channels)

        self.connect_message()

    def ask(self, cmd: str) -> str:
        ret = self.ask_raw(cmd)
        ret = ret.strip()
        return ret

    def get_idn(self) -> Dict[str, Optional[str]]:
        fw = self.ask('FIRMWARE?')
        MN = self.ask('MN?')
        SN = self.ask('SN?')

        id_dict: Dict[str, Optional[str]] = {
            'firmware': fw,
            'model': MN[3:],
            'serial': SN[3:],
            'vendor': 'Mini-Circuits'
        }
        return id_dict



import json

import numpy
import pytest

from qcodes.instrument.parameter import ManualParameter
from qcodes.tests.instrument_mocks import DummyInstrument
from qcodes.dataset.measurements import Measurement
from qcodes.station import Station

# pylint: disable=unused-import
from qcodes.tests.test_station import set_default_station_to_none


@pytest.fixture  # scope is "function" per default
def dac():
    dac = DummyInstrument('dummy_dac', gates=['ch1', 'ch2'])
    yield dac
    dac.close()


@pytest.fixture
def dmm():
    dmm = DummyInstrument('dummy_dmm', gates=['v1', 'v2'])
    yield dmm
    dmm.close()


@pytest.mark.parametrize("pass_station", (True, False))
@pytest.mark.usefixtures('set_default_station_to_none')
def test_station_snapshot_during_measurement(experiment, dac, dmm,
                                             pass_station):
    station = Station()
    station.add_component(dac)
    station.add_component(dmm, 'renamed_dmm')

    snapshot_of_station = station.snapshot()

    if pass_station:
        measurement = Measurement(experiment, station)
    else:
        # in this branch of the `if` we expect that `Measurement` object
        # will be initialized with `Station.default` which is equal to the
        # station object that is instantiated above
        measurement = Measurement(experiment)

    measurement.register_parameter(dac.ch1)
    measurement.register_parameter(dmm.v1, setpoints=[dac.ch1])

    with measurement.run() as data_saver:
        data_saver.add_result((dac.ch1, 7), (dmm.v1, 5))

    # 1. Test `get_metadata('snapshot')` method

    json_snapshot_from_dataset = data_saver.dataset.get_metadata('snapshot')
    snapshot_from_dataset = json.loads(json_snapshot_from_dataset)

    expected_snapshot = {'station': snapshot_of_station}
    assert expected_snapshot == snapshot_from_dataset

    # 2. Test `snapshot_raw` property

    assert json_snapshot_from_dataset == data_saver.dataset.snapshot_raw

    # 3. Test `snapshot` property

    assert expected_snapshot == data_saver.dataset.snapshot


@pytest.mark.usefixtures('set_default_station_to_none')
def test_snapshot_creation_for_types_not_supported_by_builtin_json(experiment):
    """
    Test that `Measurement`/`Runner`/`DataSaver` infrastructure
    successfully dumps station snapshots in JSON format in cases when the
    snapshot contains data of types that are not supported by python builtin
    `json` module, for example, numpy scalars.
    """
    p1 = ManualParameter('p_np_int32', initial_value=numpy.int32(5))
    p2 = ManualParameter('p_np_float16', initial_value=numpy.float16(5.0))
    p3 = ManualParameter('p_np_array',
                         initial_value=numpy.meshgrid((1, 2), (3, 4)))
    p4 = ManualParameter('p_np_bool', initial_value=numpy.bool_(False))

    station = Station(p1, p2, p3, p4)

    measurement = Measurement(experiment, station)

    # we need at least 1 parameter to be able to run the measurement
    measurement.register_custom_parameter('dummy')

    with measurement.run() as data_saver:
        # we do this in order to create a snapshot of the station and add it
        # to the database
        pass

    snapshot = data_saver.dataset.snapshot

    assert 5 == snapshot['station']['parameters']['p_np_int32']['value']
    assert 5 == snapshot['station']['parameters']['p_np_int32']['raw_value']

    assert 5.0 == snapshot['station']['parameters']['p_np_float16']['value']
    assert 5.0 == snapshot['station']['parameters']['p_np_float16']['raw_value']

    lst = [[[1, 2], [1, 2]], [[3, 3], [4, 4]]]
    assert lst == snapshot['station']['parameters']['p_np_array']['value']
    assert lst == snapshot['station']['parameters']['p_np_array']['raw_value']

    assert False is snapshot['station']['parameters']['p_np_bool']['value']
    assert False is snapshot['station']['parameters']['p_np_bool']['raw_value']



"""
This module defines constants that are used in Alazar ATS API.

Since the original names of the constants are mostly preserved, it is
convenient to find the uesful constants here based on the Alazar SDK manual.
"""


from typing import NewType, Dict, Tuple
from enum import IntEnum, IntFlag


ReturnCode = NewType('ReturnCode', int)

API_SUCCESS = ReturnCode(512)
API_DMA_IN_PROGRESS = ReturnCode(518)

max_buffer_size = 64 * 1024 * 1024
# The maximum size of a single buffer
# in bytes. see docs of AlazarBeforeAsyncRead
# http://www.alazartech.com/Support/Download%20Files/ATS-SDK-Guide-7.2.3.pdf#section*.110

ERROR_CODES: Dict[ReturnCode, str] = {ReturnCode(code): msg for code, msg in {
    513: 'ApiFailed',
    514: 'ApiAccessDenied',
    515: 'ApiDmaChannelUnavailable',
    516: 'ApiDmaChannelInvalid',
    517: 'ApiDmaChannelTypeError',
    518: 'ApiDmaInProgress',
    519: 'ApiDmaDone',
    520: 'ApiDmaPaused',
    521: 'ApiDmaNotPaused',
    522: 'ApiDmaCommandInvalid',
    523: 'ApiDmaManReady',
    524: 'ApiDmaManNotReady',
    525: 'ApiDmaInvalidChannelPriority',
    526: 'ApiDmaManCorrupted',
    527: 'ApiDmaInvalidElementIndex',
    528: 'ApiDmaNoMoreElements',
    529: 'ApiDmaSglInvalid',
    530: 'ApiDmaSglQueueFull',
    531: 'ApiNullParam',
    532: 'ApiInvalidBusIndex',
    533: 'ApiUnsupportedFunction',
    534: 'ApiInvalidPciSpace',
    535: 'ApiInvalidIopSpace',
    536: 'ApiInvalidSize',
    537: 'ApiInvalidAddress',
    538: 'ApiInvalidAccessType',
    539: 'ApiInvalidIndex',
    540: 'ApiMuNotReady',
    541: 'ApiMuFifoEmpty',
    542: 'ApiMuFifoFull',
    543: 'ApiInvalidRegister',
    544: 'ApiDoorbellClearFailed',
    545: 'ApiInvalidUserPin',
    546: 'ApiInvalidUserState',
    547: 'ApiEepromNotPresent',
    548: 'ApiEepromTypeNotSupported',
    549: 'ApiEepromBlank',
    550: 'ApiConfigAccessFailed',
    551: 'ApiInvalidDeviceInfo',
    552: 'ApiNoActiveDriver',
    553: 'ApiInsufficientResources',
    554: 'ApiObjectAlreadyAllocated',
    555: 'ApiAlreadyInitialized',
    556: 'ApiNotInitialized',
    557: 'ApiBadConfigRegEndianMode',
    558: 'ApiInvalidPowerState',
    559: 'ApiPowerDown',
    560: 'ApiFlybyNotSupported',
    561: 'ApiNotSupportThisChannel',
    562: 'ApiNoAction',
    563: 'ApiHSNotSupported',
    564: 'ApiVPDNotSupported',
    565: 'ApiVpdNotEnabled',
    566: 'ApiNoMoreCap',
    567: 'ApiInvalidOffset',
    568: 'ApiBadPinDirection',
    569: 'ApiPciTimeout',
    570: 'ApiDmaChannelClosed',
    571: 'ApiDmaChannelError',
    572: 'ApiInvalidHandle',
    573: 'ApiBufferNotReady',
    574: 'ApiInvalidData',
    575: 'ApiDoNothing',
    576: 'ApiDmaSglBuildFailed',
    577: 'ApiPMNotSupported',
    578: 'ApiInvalidDriverVersion',
    579: ('ApiWaitTimeout: operation did not finish during '
          'timeout interval. Check your trigger.'),
    580: 'ApiWaitCanceled',
    581: 'ApiBufferTooSmall',
    582: ('ApiBufferOverflow: rate of acquiring data > rate of '
          'transferring data to local memory. Try reducing sample rate, '
          'reducing number of enabled channels, increasing size of each '
          'DMA buffer or increase number of DMA buffers.'),
    583: 'ApiInvalidBuffer',
    584: 'ApiInvalidRecordsPerBuffer',
    585: ('ApiDmaPending: Async I/O operation was successfully started, '
          'it will be completed when sufficient trigger events are '
          'supplied to fill the buffer.'),
    586: ('ApiLockAndProbePagesFailed:Driver or operating system was '
          'unable to prepare the specified buffer for DMA transfer. '
          'Try reducing buffer size or total number of buffers.'),
    587: 'ApiWaitAbandoned',
    588: 'ApiWaitFailed',
    589: ('ApiTransferComplete: This buffer is last in the current '
          'acquisition.'),
    590: 'ApiPllNotLocked: hardware error, contact AlazarTech',
    591: ('ApiNotSupportedInDualChannelMode:Requested number of samples '
          'per channel is too large to fit in on-board memory. Try '
          'reducing number of samples per channel, or switch to '
          'single channel mode.'),
    592: ('ApiNotSupportedInQuadChannelMode: The requested number of '
          'samples per channel is too large to fit in onboard memory. '
          'Try reducing the num'),
    593: 'ApiFileIoError: A file read or write error occurred.',
    594: ('ApiInvalidClockFrequency: The requested ADC clock frequency is '
          'not supported.'),
    595: 'ApiInvalidSkipTable',
    596: 'ApiInvalidDspModule',
    597: 'ApiDESOnlySupportedInSingleChannelMode',
    598: 'ApiInconsistentChannel',
    599: 'ApiDspFiniteRecordsPerAcquisition',
    600: 'ApiNotEnoughNptFooters',
    601: 'ApiInvalidNptFooter',
    602: ('ApiOCTIgnoreBadClockNotSupported: OCT ignore bad clock is not '
          'supported.'),
    603: ('ApiError: The requested number of records in a single-port '
          'acquisition exceeds the maximum supported by the digitizer. '
          'Use dual-ported AutoDMA to acquire more records per acquisition.'),
    604: ('ApiError: The requested number of records in a single-port '
          'acquisition exceeds the maximum supported by the digitizer.'),
    605: ('ApiOCTNoTriggerDetected: No trigger is detected for OCT ignore '
          'bad clock feature.'),
    606: ('ApiOCTTriggerTooFast: Trigger detected is too fast for OCT ignore '
          'bad clock feature.'),
    607: ('ApiNetworkError: There was an issue related to network. Make sure '
          'that the network connection and settings are correct.'),
    608: ('ApiFftSizeTooLarge: The on-FPGA FFT cannot support FFT that '
          'large. Try reducing the FFT size.'),
}.items()}


BOARD_NAMES = {
    0: 'ATS_NONE',
    1: 'ATS850',
    2: 'ATS310',
    3: 'ATS330',
    4: 'ATS855',
    5: 'ATS315',
    6: 'ATS335',
    7: 'ATS460',
    8: 'ATS860',
    9: 'ATS660',
    10: 'ATS665',
    11: 'ATS9462',
    12: 'ATS9434',
    13: 'ATS9870',
    14: 'ATS9350',
    15: 'ATS9325',
    16: 'ATS9440',
    17: 'ATS9410',
    18: 'ATS9351',
    19: 'ATS9310',
    20: 'ATS9461',
    21: 'ATS9850',
    22: 'ATS9625',
    23: 'ATG6500',
    24: 'ATS9626',
    25: 'ATS9360',
    26: 'AXI9870',
    27: 'ATS9370',
    28: 'ATU7825',
    29: 'ATS9373',
    30: 'ATS9416',
    31: 'ATS9637',
    32: 'ATS9120',
    33: 'ATS9371',
    34: 'ATS9130',
    35: 'ATS9352',
    36: 'ATS9453',
}


# See table 4 Input Channel Configurations
# on page 241 of the Alazar SDK manual
_NUMBER_OF_CHANNELS_TO_BYTE_REPR: Dict[int, Tuple[int, ...]] = {
    1:  tuple(2**i for i in range(16)),
    2:  (3, 5, 6, 9, 10, 12),
    4:  (15,),
    8:  (255,),
    16: (65535,)
    }

# See table 4 Input Channel Configurations
# on page 241 of the Alazar SDK manual
NUMBER_OF_CHANNELS_FROM_BYTE_REPR: Dict[int, int] = {
    byte_repr: n_ch
        for n_ch, byte_reprs in _NUMBER_OF_CHANNELS_TO_BYTE_REPR.items()
        for byte_repr in byte_reprs
    }


REGISTER_ACCESS_PASSWORD = 0x32145876


class ClockSource(IntEnum):
    INTERNAL_CLOCK = 0x00000001
    FAST_EXTERNAL_CLOCK = 0x00000002
    EXTERNAL_CLOCK = FAST_EXTERNAL_CLOCK
    MEDIUM_EXTERNAL_CLOCK = 0x00000003
    SLOW_EXTERNAL_CLOCK = 0x00000004
    EXTERNAL_CLOCK_AC = 0x00000005
    EXTERNAL_CLOCK_DC = 0x00000006
    EXTERNAL_CLOCK_10MHz_REF = 0x00000007
    INTERNAL_CLOCK_10MHz_REF = 0x00000008
    EXTERNAL_CLOCK_10MHz_PXI = 0x0000000A
    INTERNAL_CLOCK_DIV_4 = 0x0000000F
    INTERNAL_CLOCK_DIV_5 = 0x00000010
    MASTER_CLOCK = 0x00000011
    INTERNAL_CLOCK_SET_VCO = 0x00000012


class ClockEdge(IntEnum):
    CLOCK_EDGE_RISING = 0x00000000
    CLOCK_EDGE_FALLING = 0x00000001


class SampleRate(IntEnum):
    SAMPLE_RATE_USER_DEF = 0x00000040
    SAMPLE_RATE_1KSPS = 0X00000001
    SAMPLE_RATE_2KSPS = 0X00000002
    SAMPLE_RATE_5KSPS = 0X00000004
    SAMPLE_RATE_10KSPS = 0X00000008
    SAMPLE_RATE_20KSPS = 0X0000000A
    SAMPLE_RATE_50KSPS = 0X0000000C
    SAMPLE_RATE_100KSPS = 0X0000000E
    SAMPLE_RATE_200KSPS = 0X00000010
    SAMPLE_RATE_500KSPS = 0X00000012
    SAMPLE_RATE_1MSPS = 0X00000014
    SAMPLE_RATE_2MSPS = 0X00000018
    SAMPLE_RATE_5MSPS = 0X0000001A
    SAMPLE_RATE_10MSPS = 0X0000001C
    SAMPLE_RATE_20MSPS = 0X0000001E
    SAMPLE_RATE_25MSPS = 0X00000021
    SAMPLE_RATE_50MSPS = 0X00000022
    SAMPLE_RATE_100MSPS = 0X00000024
    SAMPLE_RATE_125MSPS = 0x00000025
    SAMPLE_RATE_160MSPS = 0x00000026
    SAMPLE_RATE_180MSPS = 0x00000027
    SAMPLE_RATE_200MSPS = 0X00000028
    SAMPLE_RATE_250MSPS = 0X0000002B
    SAMPLE_RATE_400MSPS = 0X0000002D
    SAMPLE_RATE_500MSPS = 0X00000030
    SAMPLE_RATE_800MSPS = 0X00000032
    SAMPLE_RATE_1000MSPS = 0x00000035
    SAMPLE_RATE_1GSPS = SAMPLE_RATE_1000MSPS
    SAMPLE_RATE_1200MSPS = 0x00000037
    SAMPLE_RATE_1500MSPS = 0x0000003A
    SAMPLE_RATE_1600MSPS = 0x0000003B
    SAMPLE_RATE_1800MSPS = 0x0000003D
    SAMPLE_RATE_2000MSPS = 0x0000003F
    SAMPLE_RATE_2GSPS = SAMPLE_RATE_2000MSPS
    SAMPLE_RATE_2400MSPS = 0x0000006A
    SAMPLE_RATE_3000MSPS = 0x00000075
    SAMPLE_RATE_3GSPS = SAMPLE_RATE_3000MSPS
    SAMPLE_RATE_3600MSPS = 0x0000007B
    SAMPLE_RATE_4000MSPS = 0x00000080
    SAMPLE_RATE_4GSPS = SAMPLE_RATE_4000MSPS


class Coupling(IntEnum):
    AC_COUPLING = 0x00000001
    DC_COUPLING = 0x00000002


class InputRange(IntEnum):
    INPUT_RANGE_PM_20_MV = 0x00000001
    INPUT_RANGE_PM_40_MV = 0x00000002
    INPUT_RANGE_PM_50_MV = 0x00000003
    INPUT_RANGE_PM_80_MV = 0x00000004
    INPUT_RANGE_PM_100_MV = 0x00000005
    INPUT_RANGE_PM_200_MV = 0x00000006
    INPUT_RANGE_PM_400_MV = 0x00000007
    INPUT_RANGE_PM_500_MV = 0x00000008
    INPUT_RANGE_PM_800_MV = 0x00000009
    INPUT_RANGE_PM_1_V = 0x0000000A
    INPUT_RANGE_PM_2_V = 0x0000000B
    INPUT_RANGE_PM_4_V = 0x0000000C
    INPUT_RANGE_PM_5_V = 0x0000000D
    INPUT_RANGE_PM_8_V = 0x0000000E
    INPUT_RANGE_PM_10_V = 0x0000000F
    INPUT_RANGE_PM_20_V = 0x00000010
    INPUT_RANGE_PM_40_V = 0x00000011
    INPUT_RANGE_PM_16_V = 0x00000012
    INPUT_RANGE_HIFI = 0x00000020
    INPUT_RANGE_PM_1_V_25 = 0x00000021
    INPUT_RANGE_PM_2_V_5 = 0x00000025
    INPUT_RANGE_PM_125_MV = 0x00000028
    INPUT_RANGE_PM_250_MV = 0x00000030


class Impedance(IntEnum):
    IMPEDANCE_1M_OHM = 0x00000001
    IMPEDANCE_50_OHM = 0x00000002
    IMPEDANCE_75_OHM = 0x00000004
    IMPEDANCE_300_OHM = 0x00000008


class BandwidthLimit(IntEnum):
    DISABLE = 0
    ENABLE = 1


class TriggerEngine(IntEnum):
    TRIG_ENGINE_J = 0x00000000
    TRIG_ENGINE_K = 0x00000001


class TriggerEngineOperation(IntEnum):
    TRIG_ENGINE_OP_J = 0x00000000
    TRIG_ENGINE_OP_K = 0x00000001
    TRIG_ENGINE_OP_J_OR_K = 0x00000002
    TRIG_ENGINE_OP_J_AND_K = 0x00000003
    TRIG_ENGINE_OP_J_XOR_K = 0x00000004
    TRIG_ENGINE_OP_J_AND_NOT_K = 0x00000005
    TRIG_ENGINE_OP_NOT_J_AND_K = 0x00000006


class TriggerEngineSource(IntEnum):
    TRIG_CHAN_A = 0x00000000
    TRIG_CHAN_B = 0x00000001
    TRIG_EXTERNAL = 0x00000002
    TRIG_DISABLE = 0x00000003
    TRIG_CHAN_C = 0x00000004
    TRIG_CHAN_D = 0x00000005
    TRIG_CHAN_E = 0x00000006
    TRIG_CHAN_F = 0x00000007
    TRIG_CHAN_G = 0x00000008
    TRIG_CHAN_H = 0x00000009
    TRIG_CHAN_I = 0x0000000A
    TRIG_CHAN_J = 0x0000000B
    TRIG_CHAN_K = 0x0000000C
    TRIG_CHAN_L = 0x0000000D
    TRIG_CHAN_M = 0x0000000E
    TRIG_CHAN_N = 0x0000000F
    TRIG_CHAN_O = 0x00000010
    TRIG_CHAN_P = 0x00000011
    TRIG_PXI_STAR = 0x00000100


class TriggerSlope(IntEnum):
    """
    Used in more than one place, for example, for ``AUX_IN_TRIGGER_ENABLE``
    Auxiallary IO mode
    """
    TRIGGER_SLOPE_POSITIVE = 0x00000001
    TRIGGER_SLOPE_NEGATIVE = 0x00000002


class ExternalTriggerCoupling(IntEnum):
    AC = 1
    DC = 2


class ExternalTriggerRange(IntEnum):
    ETR_DIV5 = 0x00000000
    ETR_X1 = 0x00000001
    ETR_5V = 0x00000000
    ETR_1V = 0x00000001
    ETR_TTL = 0x00000002
    ETR_2V5 = 0x00000003


# This is how it's called in AlazarCmd.h
ExternalTriggerAttenuatorRelay = ExternalTriggerRange


class AuxilaryIO(IntEnum):
    # Output
    AUX_OUT_TRIGGER = 0
    AUX_OUT_PACER = 2
    AUX_OUT_BUSY = 4
    AUX_OUT_CLOCK = 6
    AUX_OUT_RESERVED = 8
    AUX_OUT_CAPTURE_ALMOST_DONE = 10
    AUX_OUT_AUXILIARY = 12
    AUX_OUT_SERIAL_DATA = 14
    AUX_OUT_TRIGGER_ENABLE = 16
    # Input
    AUX_IN_TRIGGER_ENABLE = 1
    AUX_IN_DIGITAL_TRIGGER = 3
    AUX_IN_GATE = 5
    AUX_IN_CAPTURE_ON_DEMAND = 7
    AUX_IN_RESET_TIMESTAMP = 9
    AUX_IN_SLOW_EXTERNAL_CLOCK = 11
    AUX_IN_AUXILIARY = 13
    AUX_IN_SERIAL_DATA = 15


class AutoDMAFlag(IntFlag):
    # Automatic DMA acquisition modes
    ADMA_TRADITIONAL_MODE = 0x00000000
    ADMA_CONTINUOUS_MODE = 0x00000100
    ADMA_NPT = 0x00000200
    ADMA_TRIGGERED_STREAMING = 0x00000400
    # Automatic DMA acquisition options
    ADMA_EXTERNAL_STARTCAPTURE = 0x00000001
    ADMA_ENABLE_RECORD_HEADERS = 0x00000008
    ADMA_FIFO_ONLY_STREAMING = 0x00000800
    ADMA_ALLOC_BUFFERS = 0x00000020
    ADMA_INTERLEAVE_SAMPLES = 0x00001000
    ADMA_GET_PROCESSED_DATA = 0x00002000
    ADMA_DSP = 0x00004000
    # Other
    ADMA_SINGLE_DMA_CHANNEL = 0x00000010
    ADMA_ENABLE_RECORD_FOOTERS = 0x00010000


class Channel(IntFlag):
    """Flags for selecting channel configuration"""
    ALL = 0x00000000
    A = 0x00000001
    B = 0x00000002
    C = 0x00000004
    D = 0x00000008
    E = 0x00000010
    F = 0x00000020
    G = 0x00000040
    H = 0x00000080
    I = 0x00000100
    J = 0x00000200
    K = 0x00000400
    L = 0x00000800
    M = 0x00001000
    N = 0x00002000
    O = 0x00004000
    P = 0x00008000


class Capability(IntEnum):
    """Capability identifiers for :meth:`.AlazarATSAPI.query_capability`"""
    GET_SERIAL_NUMBER = 0x10000024
    # Date of the board's latest calibration data as a decimal number with
    # the format DDMMYY where DD is 1-31, MM is 1-12,
    # and YY is 00-99 from 2000
    GET_LATEST_CAL_DATE = 0x10000026
    # Month of the board's latest calibration date as
    # a decimal number with the format MM where M is 1-12
    GET_LATEST_CAL_DATE_MONTH = 0x1000002D
    # Day of month of the board's latest calibration date
    # as a decimal number with the format DD where DD is 1-31
    GET_LATEST_CAL_DATE_DAY = 0x1000002E
    # Year of the board's latest calibration date
    # as a decimal number with the format YY where YY is 00-99 from 2000
    GET_LATEST_CAL_DATE_YEAR = 0x1000002F
    # On-board memory size in maximum samples per channel in single channel
    # mode; see AlazarGetChannelInfo for more information
    MEMORY_SIZE = 0x1000002A
    # Board's FPGA signature
    ASOPC_TYPE = 0x1000002C
    # The PCIe link speed negotiated between a PCIe digitizer board and
    # the host PCIe bus in 2.5G bits per second units. The PCIe bus uses
    # 10b/8b encoding, so divide the link speed by 10 to find the link speed
    # in bytes per second. For example, a link speed of 2.5 Gb/s gives
    # 250 MB/s per lane. PCIe Gen 2 digitizers such as ATS9360 should receive
    # 5.0 Gb/s links. PCIe Gen 3 digitizers such as ATS9373 should receive
    # 8.0 Gb/s links.
    GET_PCIE_LINK_SPEED = 0x10000030
    # The PCIe link width in lanes negociated between a PCIe digitizer board
    # and the host PCIe bus. An ATS9462 should negociate 4 lanes, while the
    # ATS9325, ATS9350, ATS9351, ATS9360, ATS9373, ATS9440, ATS9850 and
    # ATS9870 should negociate 8 lanes. If a board obtains fewer lanes, then
    # the board may be installed in a PCIe slot that does not support
    # the expected number of lanes. The ideal PCIe bandwidth is the link speed
    # in bytes per second per lane, multiplied by the link width in lanes.
    # For example, and ATS9870 that negociates 8 lanes at 250 MB/s per lane
    # has an ideal bandwidth of 2 GB/s.
    GET_PCIE_LINK_WIDTH = 0x10000031
    # Board type identifier; see AlazarGetBoardKind for more information
    BOARD_TYPE = 0x1000002B
    # Get the maximum number of pre-trigger samples
    GET_MAX_PRETRIGGER_SAMPLES = 0x10000046
    # The model of user-programable FPGA device (see also `CPFDevice`)
    GET_CPF_DEVICE = 0x10000071


CPF_DEVICE_EP3SL50 = 1
CPF_DEVICE_EP3SE260 = 2


class CPFDevice(IntEnum):
    EP3SL50 = CPF_DEVICE_EP3SL50
    EP3SE260 = CPF_DEVICE_EP3SE260


CRA_MODE_DISABLE = 0
CRA_MODE_ENABLE_FPGA_AVE = 1


class RecordAverageMode(IntEnum):
    """
    Values for ``mode`` argument of ``AlazarConfigureRecordAverage`` function
    """
    DISABLE = CRA_MODE_DISABLE
    ENABLE_FPGA_AVE = CRA_MODE_ENABLE_FPGA_AVE


CRA_OPTION_UNSIGNED = 0
CRA_OPTION_SIGNED = 1


class RecordAverageOption(IntEnum):
    """
    Values for ``options`` argument of ``AlazarConfigureRecordAverage``
    function
    """
    UNSIGNED = CRA_OPTION_UNSIGNED  #: Find sum of unsigned ADC samples codes
    SIGNED = CRA_OPTION_SIGNED  #: Find sum of signed ADC sample


LED_OFF = 0
LED_ON = 1


class AlazarParameter(IntEnum):
    """
    Parameters suitable to be used with `` AlazarSetParameter`` and/or
    ``AlazarGetParameter``
    Defined by ``ALAZAR_PARAMETERS`` in ``AlazarCmd.h``
    """
    DATA_WIDTH = 0x10000009
    # The number of bits per sample
    SETGET_ASYNC_BUFFSIZE_BYTES = 0x10000039
    # The size of API-allocated DMA buffers in bytes
    SETGET_ASYNC_BUFFCOUNT = 0x10000040
    # The number of API-allocated DMA buffers
    GET_ASYNC_BUFFERS_PENDING = 0x10000050
    # DMA buffers currently posted to the board
    GET_ASYNC_BUFFERS_PENDING_FULL = 0x10000051
    # DMA buffers waiting to be processed by the application
    GET_ASYNC_BUFFERS_PENDING_EMPTY = 0x10000052
    # DMA buffers waiting to be filled by the board
    SET_DATA_FORMAT = 0x10000041
    # 0 if the data format is unsigned, and 1 otherwise
    GET_DATA_FORMAT = 0x10000042
    # 0 if the data format is unsigned, and 1 otherwise
    GET_SAMPLES_PER_TIMESTAMP_CLOCK = 0x10000044
    # Number of samples per timestamp clock
    GET_RECORDS_CAPTURED = 0x10000045
    # Records captured since the start of the acquisition (single-port)
    # or buffer (dual-port)
    ECC_MODE = 0x10000048
    # ECC mode. Member of ECCMode enum
    GET_AUX_INPUT_LEVEL = 0x10000049
    # Read the TTL level of the AUX connector.
    # Member of  AUXInputLevel enum
    GET_CHANNELS_PER_BOARD = 0x10000070
    # Number of analog channels supported by this digitizer
    GET_FPGA_TEMPERATURE = 0x10000080
    # Current FPGA temperature in degrees Celcius. Only supported by
    # PCIe digitizers.
    PACK_MODE = 0x10000072
    # Get/Set the pack mode as a member of PackMode enum
    SET_SINGLE_CHANNEL_MODE = 0x10000043
    # Reserve all the on-board memory to the channel passed as
    # argument. Single-port only.
    API_FLAGS = 0x10000090
    # State of the API logging as a member of
    # API_TRACE_STATES enum


class ECCMode(IntEnum):
    """
    Values for ECC_MODE of ``Parameter``
    Defined by ``ALAZAR_ECC_MODES`` in ``AlazarCmd.h``
    """
    ECC_DISABLE = 0  # Disable
    ECC_ENABLE = 1  # Enable


class PackMode(IntEnum):
    """
    Values for PACK_MODE of ``Parameter``
    Defined by ``ALAZAR_PACK_MODES`` in ``AlazarCmd.h``
    """
    PACK_DEFAULT = 0  # Default pack mode of the board
    PACK_8_BITS_PER_SAMPLE = 1  # 8 bits per sample
    PACK_12_BITS_PER_SAMPLE = 2  # 12 bits per sample


class AUXInputLevel(IntEnum):
    """
    Values for GET_AUX_INPUT_LEVEL of ``Parameter``
    Defined by ``ALAZAR_AUX_INPUT_LEVELS`` in ``AlazarCmd.h``
    """
    AUX_INPUT_LOW = 0  # Low level
    AUX_INPUT_HIGH = 1  # High level


class APITraceStates(IntEnum):
    """
    Values for API_FLAGS of ``Parameter``
    Defined by ``ALAZAR_API_TRACE_STATES`` in ``AlazarCmd.h``
    """
    API_ENABLE_TRACE = 1  # Trace enabled
    API_DISABLE_TRACE = 0  # Trace disabled



from collections import Counter

import numpy as np
import pytest

import qcodes.instrument.sims as sims
from qcodes.instrument_drivers.tektronix.Keithley_2600_channels import (
    Keithley_2600,
    MeasurementStatus,
)

visalib = sims.__file__.replace('__init__.py', 'Keithley_2600.yaml@sim')


@pytest.fixture(scope='function')
def driver():
    driver = Keithley_2600('Keithley_2600',
                           address='GPIB::1::INSTR',
                           visalib=visalib)

    yield driver
    driver.close()


@pytest.fixture(scope='function')
def smus(driver):
    smu_names = {'smua', 'smub'}
    assert smu_names == set(list(driver.submodules.keys()))

    yield tuple(getattr(driver, smu_name)
                for smu_name in smu_names)


def test_idn(driver):
    assert {'firmware': '3.0.0',
            'model': '2601B',
            'serial': '1398687',
            'vendor': 'Keithley Instruments Inc.'} == driver.IDN()


def test_smu_channels_and_their_parameters(driver):
    assert {'smua', 'smub'} == set(list(driver.submodules.keys()))

    for smu_name in {'smua', 'smub'}:
        smu = getattr(driver, smu_name)

        smu.volt(1.0)
        assert smu.volt.measurement_status == None

        assert 1.0 == smu.volt()
        assert smu.volt.measurement_status == MeasurementStatus.NORMAL

        smu.curr(1.0)
        assert smu.volt.measurement_status == None

        assert 1.0 == smu.curr()
        assert smu.curr.measurement_status == MeasurementStatus.NORMAL

        assert 0.0 == smu.res()

        assert 'current' == smu.mode()
        smu.mode('voltage')
        assert smu.mode() == 'voltage'

        assert smu.output() is False
        smu.output(True)
        assert smu.output() is True

        assert 0.0 == smu.nplc()
        smu.nplc(2.3)
        assert smu.nplc() == 2.3

        assert 0.0 == smu.sourcerange_v()
        some_valid_sourcerange_v = driver._vranges[smu.model][2]
        smu.sourcerange_v(some_valid_sourcerange_v)
        assert smu.sourcerange_v() == some_valid_sourcerange_v

        assert smu.source_autorange_v_enabled() is False
        smu.source_autorange_v_enabled(True)
        assert smu.source_autorange_v_enabled() is True

        assert 0.0 == smu.measurerange_v()
        some_valid_measurerange_v = driver._vranges[smu.model][2]
        smu.measurerange_v(some_valid_measurerange_v)
        assert smu.measurerange_v() == some_valid_measurerange_v

        assert smu.measure_autorange_v_enabled() is False
        smu.measure_autorange_v_enabled(True)
        assert smu.measure_autorange_v_enabled() is True

        assert 0.0 == smu.sourcerange_i()
        some_valid_sourcerange_i = driver._iranges[smu.model][2]
        smu.sourcerange_i(some_valid_sourcerange_i)
        assert smu.sourcerange_i() == some_valid_sourcerange_i

        assert smu.source_autorange_i_enabled() is False
        smu.source_autorange_i_enabled(True)
        assert smu.source_autorange_i_enabled() is True

        assert 0.0 == smu.measurerange_i()
        some_valid_measurerange_i = driver._iranges[smu.model][2]
        smu.measurerange_i(some_valid_measurerange_i)
        assert smu.measurerange_i() == some_valid_measurerange_i

        assert smu.measure_autorange_i_enabled() is False
        smu.measure_autorange_i_enabled(True)
        assert smu.measure_autorange_i_enabled() is True

        assert 0.0 == smu.limitv()
        smu.limitv(2.3)
        assert smu.limitv() == 2.3

        assert 0.0 == smu.limiti()
        smu.limiti(2.3)
        assert smu.limiti() == 2.3

        assert 'current' == smu.timetrace_mode()
        smu.timetrace_mode('voltage')
        assert smu.timetrace_mode() == 'voltage'

        assert 500 == smu.timetrace_npts()
        smu.timetrace_npts(600)
        assert smu.timetrace_npts() == 600

        assert 0.001 == smu.timetrace_dt()
        smu.timetrace_dt(0.002)
        assert smu.timetrace_dt() == 0.002

        dt = smu.timetrace_dt()
        npts = smu.timetrace_npts()
        expected_time_axis = np.linspace(0, dt*npts, npts, endpoint=False)
        assert len(expected_time_axis) == len(smu.time_axis())
        assert Counter(expected_time_axis) == Counter(smu.time_axis())
        assert set(expected_time_axis) == set(smu.time_axis())

        smu.timetrace_mode('current')
        assert 'A' == smu.timetrace.unit
        assert 'Current' == smu.timetrace.label
        assert smu.time_axis == smu.timetrace.setpoints[0]

        smu.timetrace_mode('voltage')
        assert 'V' == smu.timetrace.unit
        assert 'Voltage' == smu.timetrace.label
        assert smu.time_axis == smu.timetrace.setpoints[0]


def test_setting_source_voltage_range_disables_autorange(smus):
    for smu in smus:
        smu.source_autorange_v_enabled(True)
        assert smu.source_autorange_v_enabled() is True
        some_valid_sourcerange_v = smu.root_instrument._vranges[smu.model][2]
        smu.sourcerange_v(some_valid_sourcerange_v)
        assert smu.source_autorange_v_enabled() is False


def test_setting_measure_voltage_range_disables_autorange(smus):
    for smu in smus:
        smu.measure_autorange_v_enabled(True)
        assert smu.measure_autorange_v_enabled() is True
        some_valid_measurerange_v = smu.root_instrument._vranges[smu.model][2]
        smu.measurerange_v(some_valid_measurerange_v)
        assert smu.measure_autorange_v_enabled() is False


def test_setting_source_current_range_disables_autorange(smus):
    for smu in smus:
        smu.source_autorange_i_enabled(True)
        assert smu.source_autorange_i_enabled() is True
        some_valid_sourcerange_i = smu.root_instrument._iranges[smu.model][2]
        smu.sourcerange_i(some_valid_sourcerange_i)
        assert smu.source_autorange_i_enabled() is False


def test_setting_measure_current_range_disables_autorange(smus):
    for smu in smus:
        smu.measure_autorange_i_enabled(True)
        assert smu.measure_autorange_i_enabled() is True
        some_valid_measurerange_i = smu.root_instrument._iranges[smu.model][2]
        smu.measurerange_i(some_valid_measurerange_i)
        assert smu.measure_autorange_i_enabled() is False



"""
This module defines the dict representation of the ``RunDescriber`` used for
serialization and deserialization of the ``RunDescriber``.

RunDescriber version log:

- 0: The run_describer has a single attribute, interdependencies, which is an
instance of InterDependencies (which contains ParamSpecs)
- 1: The run_describer has a single attribute, interdependencies, which is an
instance of InterDependencies_ (which contains ParamSpecBases)
- 2: The run_describer has a two attributes: interdependencies, which is an
instance of InterDependencies (which contains ParamSpecs) and
interdependencies_, which is an instance of InterDependencies_
(which contains ParamSpecBases)
"""
from typing import Dict, List, Optional, Tuple, Union

from typing_extensions import TypedDict

from ..param_spec import ParamSpecBaseDict, ParamSpecDict


class InterDependenciesDict(TypedDict):
    paramspecs: Tuple[ParamSpecDict, ...]


class InterDependencies_Dict(TypedDict):
    parameters: Dict[str, ParamSpecBaseDict]
    dependencies: Dict[str, List[str]]
    inferences: Dict[str, List[str]]
    standalones: List[str]


Shapes = Optional[Dict[str, Tuple[int, ...]]]


class RunDescriberV0Dict(TypedDict):
    version: int
    interdependencies: "InterDependenciesDict"


class RunDescriberV1Dict(TypedDict):
    version: int
    interdependencies: "InterDependencies_Dict"


class RunDescriberV2Dict(RunDescriberV0Dict):
    interdependencies_: "InterDependencies_Dict"


class RunDescriberV3Dict(RunDescriberV2Dict):
    shapes: Shapes
    # dict from dependent to dict from depenency to num points in grid


RunDescriberDicts = Union[RunDescriberV0Dict,
                          RunDescriberV1Dict,
                          RunDescriberV2Dict,
                          RunDescriberV3Dict]



import functools
import logging
import time
from queue import Empty, Queue
from threading import Thread
from typing import TYPE_CHECKING, Any, Callable, List, Mapping, Optional

from qcodes.dataset.sqlite.connection import atomic_transaction

if TYPE_CHECKING:
    from qcodes.dataset.data_set import DataSet


class _Subscriber(Thread):
    """
    Class to add a subscriber to a :class:`.DataSet`. The subscriber gets called every
    time an insert is made to the results_table.

    The _Subscriber is not meant to be instantiated directly, but rather used
    via the 'subscribe' method of the :class:`.DataSet`.

    NOTE: A subscriber should be added *after* all parameters have been added.

    NOTE: Special care shall be taken when using the *state* object: it is the
    user's responsibility to operate with it in a thread-safe way.
    """

    def __init__(self,
                 dataSet: 'DataSet',
                 id_: str,
                 callback: Callable[..., None],
                 state: Optional[Any] = None,
                 loop_sleep_time: int = 0,  # in milliseconds
                 min_queue_length: int = 1,
                 callback_kwargs: Optional[Mapping[str, Any]] = None
                 ) -> None:
        super().__init__()

        self._id = id_

        self.dataSet = dataSet
        self.table_name = dataSet.table_name
        self._data_set_len = len(dataSet)

        self.state = state

        self.data_queue: "Queue[Any]" = Queue()
        self._queue_length: int = 0
        self._stop_signal: bool = False
        # convert milliseconds to seconds
        self._loop_sleep_time = loop_sleep_time / 1000
        self.min_queue_length = min_queue_length

        if callback_kwargs is None or len(callback_kwargs) == 0:
            self.callback = callback
        else:
            self.callback = functools.partial(callback, **callback_kwargs)

        self.callback_id = f"callback{self._id}"
        self.trigger_id = f"sub{self._id}"

        conn = dataSet.conn

        conn.create_function(self.callback_id, -1, self._cache_data_to_queue)

        parameters = dataSet.get_parameters()
        sql_param_list = ",".join(f"NEW.{p.name}" for p in parameters)
        sql_create_trigger_for_callback = f"""
        CREATE TRIGGER {self.trigger_id}
            AFTER INSERT ON '{self.table_name}'
        BEGIN
            SELECT {self.callback_id}({sql_param_list});
        END;"""
        atomic_transaction(conn, sql_create_trigger_for_callback)

        self.log = logging.getLogger(f"_Subscriber {self._id}")

    def _cache_data_to_queue(self, *args: Any) -> None:
        self.data_queue.put(args)
        self._data_set_len += 1
        self._queue_length += 1

    def run(self) -> None:
        self.log.debug("Starting subscriber")
        self._loop()

    @staticmethod
    def _exhaust_queue(queue: "Queue[Any]") -> List[Any]:
        result_list = []
        while True:
            try:
                result_list.append(queue.get(block=False))
            except Empty:
                break
        return result_list

    def _call_callback_on_queue_data(self) -> None:
        result_list = self._exhaust_queue(self.data_queue)
        self.callback(result_list, self._data_set_len, self.state)

    def _loop(self) -> None:
        while True:
            if self._stop_signal:
                self._clean_up()
                break

            if self._queue_length >= self.min_queue_length:
                self._call_callback_on_queue_data()
                self._queue_length = 0

            time.sleep(self._loop_sleep_time)

            if self.dataSet.completed:
                self._call_callback_on_queue_data()
                break

    def done_callback(self) -> None:
        self._call_callback_on_queue_data()

    def schedule_stop(self) -> None:
        if not self._stop_signal:
            self.log.debug("Scheduling stop")
            self._stop_signal = True

    def _clean_up(self) -> None:
        self.log.debug("Stopped subscriber")



import json
import warnings
from collections import OrderedDict, UserDict

import numpy as np
import pytest

with warnings.catch_warnings():
    # this context manager can be removed when uncertainties
    # no longer triggers deprecation warnings
    warnings.simplefilter("ignore", category=DeprecationWarning)
    import uncertainties

from qcodes.utils.helpers import NumpyJSONEncoder
from qcodes.utils.types import numpy_complex, numpy_floats, numpy_ints


def test_python_types():
    e = NumpyJSONEncoder()

    # test basic python types
    od = OrderedDict()
    od['a'] = 0
    od['b'] = 1
    testinput = [None, True, False, 10, float(10.), 'hello',
                 od]
    testoutput = ['null', 'true', 'false', '10', '10.0', '"hello"',
                  '{"a": 0, "b": 1}']

    for d, r in zip(testinput, testoutput):
        v = e.encode(d)
        if isinstance(d, dict):
            assert v == r
        else:
            assert v == r


def test_complex_types():
    e = NumpyJSONEncoder()
    assert e.encode(complex(1, 2)) == \
           '{"__dtype__": "complex", "re": 1.0, "im": 2.0}'
    for complex_type in numpy_complex:
        assert e.encode(complex_type(complex(1, 2))) == \
               '{"__dtype__": "complex", "re": 1.0, "im": 2.0}'


def test_UFloat_type():
    e = NumpyJSONEncoder()
    assert e.encode(uncertainties.ufloat(1.0, 2.0)) == \
           '{"__dtype__": "UFloat", "nominal_value": 1.0, "std_dev": 2.0}'


def test_numpy_int_types():
    e = NumpyJSONEncoder()

    for int_type in numpy_ints:
        assert e.encode(int_type(3)) == '3'


def test_numpy_float_types():
    e = NumpyJSONEncoder()

    for float_type in numpy_floats:
        assert e.encode(float_type(2.5)) == '2.5'


def test_numpy_bool_type():
    e = NumpyJSONEncoder()

    assert e.encode(np.bool_(True)) == 'true'
    assert e.encode(np.int8(5) == 5) == 'true'
    assert e.encode(np.array([8, 5]) == 5) == '[false, true]'


def test_numpy_array():
    e = NumpyJSONEncoder()

    assert e.encode(np.array([1, 0, 0])) == \
           '[1, 0, 0]'

    assert e.encode(np.arange(1.0, 3.0, 1.0)) == \
           '[1.0, 2.0]'

    assert e.encode(np.meshgrid((1, 2), (3, 4))) == \
           '[[[1, 2], [1, 2]], [[3, 3], [4, 4]]]'


def test_non_serializable():
    """
    Test that non-serializable objects are serialzed to their
    string representation
    """
    e = NumpyJSONEncoder()

    class Dummy:
        def __str__(self):
            return 'i am a dummy with \\ slashes /'

    assert e.encode(Dummy()) == \
           '"i am a dummy with \\\\ slashes /"'


def test_object_with_serialization_method():
    """
    Test that objects with `_JSONEncoder` method are serialized via
    calling that method
    """
    e = NumpyJSONEncoder()

    class Dummy:
        def __init__(self):
            self.confession = 'a_dict_addict'

        def __str__(self):
            return 'me as a string'

        def _JSONEncoder(self):
            return {'i_am_actually': self.confession}

    assert e.encode(Dummy()) == \
           '{"i_am_actually": "a_dict_addict"}'


class SomeUserDict(UserDict):
    pass


EXAMPLEMETADATA = {
    'name': 'Rapunzel',
    'age': np.int64(12),
    'height': np.float64(112.234),
    'scores': np.linspace(0, 42, num=3),
    # include some regular values to ensure they work right
    # with our encoder
    'weight': 19,
    'length': 45.23,
    'points': [12, 24, 48],
    'RapunzelNumber': np.float64(4.89) + np.float64(0.11) * 1j,
    'verisimilitude': 1j,
    'myuserdict': SomeUserDict({'a': 1})
}


def test_standard_encoder_fails_examplemetadata():
    with pytest.raises(TypeError):
        json.dumps(
            EXAMPLEMETADATA,
            sort_keys=True,
            indent=4,
            ensure_ascii=False)


def test_numpy_encoder_examplemetadata():
    data = json.dumps(EXAMPLEMETADATA, sort_keys=True, indent=4,
                      ensure_ascii=False, cls=NumpyJSONEncoder)
    data_dict = json.loads(data)

    metadata = {
        'name': 'Rapunzel',
        'age': 12,
        'height': 112.234,
        'scores': [0, 21, 42],
        'weight': 19,
        'length': 45.23,
        'points': [12, 24, 48],
        'RapunzelNumber': {'__dtype__': 'complex', 're': 4.89, 'im': 0.11},
        'verisimilitude': {'__dtype__': 'complex', 're': 0, 'im': 1},
        'myuserdict': {'a': 1}
    }
    assert metadata == data_dict



import logging
import pytest

from qcodes.instrument_drivers.tektronix.Keithley_s46 import (
    S46, LockAcquisitionError
)

import qcodes.instrument.sims as sims
visalib = sims.__file__.replace('__init__.py', 'Keithley_s46.yaml@sim')


def test_aliases_dict():
    """
    Test the class attribute 'aliases' which maps channel aliases
    (e.g. A1, B2, etc) to channel numbers.
    """
    def calc_channel_nr(alias: str) -> int:
        """
        We perform the calculation in a different way to verify correctness
        """
        offset_dict = dict(zip(["A", "B", "C", "D", "R"], range(0, 32, 6)))
        return offset_dict[alias[0]] + int(alias[1:])

    assert all([
        nr == calc_channel_nr(al) for al, nr in S46.channel_numbers.items()
    ])


@pytest.fixture(scope='module')
def s46_six():
    """
    A six channel-per-relay instrument
    """
    driver = S46('s46_six', address='GPIB::2::INSTR', visalib=visalib)

    try:
        yield driver
    finally:
        driver.close()


@pytest.fixture(scope='module')
def s46_four():
    """
    A four channel-per-relay instrument
    """
    driver = S46('s46_four', address='GPIB::3::INSTR', visalib=visalib)

    try:
        yield driver
    finally:
        driver.close()


def test_runtime_error_on_bad_init(request):
    """
    If we initialize the driver from an instrument state with more then one
    channel per relay closed, raise a runtime error. An instrument can come to
    this state if previously, other software was used to control the instrument
    """
    request.addfinalizer(S46.close_all)

    with pytest.raises(
        RuntimeError,
        match="The driver is initialized from an undesirable instrument state"
    ):
        S46('s46_bad_state', address='GPIB::1::INSTR', visalib=visalib)


def test_query_close_once_at_init(caplog):
    """
    Test that, during initialisation, we query the closed channels only once
    """
    with caplog.at_level(logging.DEBUG):
        inst = S46('s46_test_query_once', address='GPIB::2::INSTR', visalib=visalib)
        assert caplog.text.count(":CLOS?") == 1
        inst.close()


def test_init_six(s46_six, caplog):
    """
    Test that the six channel instrument initializes correctly.
    """
    assert len(s46_six.available_channels) == 26

    closed_channel_numbers = [1, 8, 13]
    assert s46_six.closed_channels() == [
        S46.aliases[i] for i in closed_channel_numbers
    ]

    with caplog.at_level(logging.DEBUG):
        s46_six.open_all_channels()
        assert ":open (@1)" in caplog.text
        assert ":open (@8)" in caplog.text
        assert ":open (@13)" in caplog.text

        assert s46_six.A1._lock._locked_by is None
        assert s46_six.B1._lock._locked_by is None
        assert s46_six.C1._lock._locked_by is None


def test_init_four(s46_four):
    """
    Test that the six channel instrument initializes correctly.
    """
    assert len(s46_four.available_channels) == 18

    closed_channel_numbers = [1, 8]
    assert s46_four.closed_channels() == [
        S46.aliases[i] for i in closed_channel_numbers
    ]

    # A four channel instrument will have channels missing
    for relay in ["A", "B", "C", "D"]:
        for index in [5, 6]:
            alias = f"{relay}{index}"
            assert not hasattr(s46_four, alias)


def test_channel_number_invariance(s46_four, s46_six):
    """
    Regardless of the channel layout (that is, number of channels per relay),
    channel aliases should represent the same channel. See also page 2-5 of the
    manual (e.g. B1 is *always* channel 7)
    """
    for alias in S46.channel_numbers.keys():
        if hasattr(s46_four, alias) and hasattr(s46_six, alias):
            channel_four = getattr(s46_four, alias)
            channel_six = getattr(s46_six, alias)
            assert channel_four.channel_number == channel_six.channel_number


def test_locking_mechanism(s46_six):
    """
    1) Test that the lock acquisition error is raised if we try to close
    more then once channel per replay
    2) Test that the lock is released when opening a channel that was closed
    """
    s46_six.A1("close")

    with pytest.raises(
        LockAcquisitionError,
        match="is already in use by channel"
    ):
        # A1 should be closed already
        s46_six.A2("close")
    # release the lock
    s46_six.A1("open")
    # now we should be able to close A2
    s46_six.A2("close")

    # Let C1 acquire the lock
    s46_six.C1("close")
    # closing C2 should raise an error
    with pytest.raises(
        LockAcquisitionError,
        match="is already in use by channel"
    ):
        s46_six.C2("close")

    # Upon opening C1 we should be able to close C2
    s46_six.C1("open")
    s46_six.C2("close")


def test_is_closed(s46_six):
    """
    Test the `is_closed` public method
    """
    assert s46_six.A1.is_closed()
    assert s46_six.B2.is_closed()
    assert s46_six.C1.is_closed()

    assert not s46_six.A2.is_closed()
    assert not s46_six.B4.is_closed()
    assert not s46_six.C6.is_closed()



"""
This module provides infrastructure for upgrading schema and structure of
QCoDeS database. It also provides concreted upgrade functions between all
the database versions which exist so far.

The module contains :mod:`.version` module for working with the version of
QCoDeS databases.

If needed, this module may contain modules with upgrade-specific code. The
intention is to be able to decouple the code of the upgrade functions from
the current state of the SQLite API in QCoDeS (:mod:`.sqlite`). In
principle, the upgrade functions should not have dependecies from
:mod:`.queries` module.
"""
import logging
import sys
from functools import wraps
from typing import Any, Callable, Dict

import numpy as np
from tqdm import tqdm

from qcodes.dataset.guids import generate_guid
from qcodes.dataset.sqlite.connection import (
    ConnectionPlus,
    atomic,
    atomic_transaction,
    transaction,
)
from qcodes.dataset.sqlite.db_upgrades.version import get_user_version, set_user_version
from qcodes.dataset.sqlite.query_helpers import insert_column, many_many, one

log = logging.getLogger(__name__)


# INFRASTRUCTURE FOR UPGRADE FUNCTIONS


TUpgraderFunction = Callable[[ConnectionPlus], None]

# Functions decorated as 'upgrader' are inserted into this dict
# The newest database version is thus determined by the number of upgrades
# in this module
# The key is the TARGET VERSION of the upgrade, i.e. the first key is 1
_UPGRADE_ACTIONS: Dict[int, Callable[..., Any]] = {}


def _latest_available_version() -> int:
    """Return latest available database schema version"""
    return len(_UPGRADE_ACTIONS)


def upgrader(func: TUpgraderFunction) -> TUpgraderFunction:
    """
    Decorator for database version upgrade functions. An upgrade function
    must have the name `perform_db_upgrade_N_to_M` where N = M-1. For
    simplicity, an upgrade function must take a single argument of type
    `ConnectionPlus`. The upgrade function must either perform the upgrade
    and return (no return values allowed) or fail to perform the upgrade,
    in which case it must raise a RuntimeError. A failed upgrade must be
    completely rolled back before the RuntimeError is raises.

    The decorator takes care of logging about the upgrade and managing the
    database versioning.
    """
    name_comps = func.__name__.split('_')
    if not len(name_comps) == 6:
        raise NameError('Decorated function not a valid upgrader. '
                        'Must have name "perform_db_upgrade_N_to_M"')
    if not ''.join(name_comps[:3]+[name_comps[4]]) == 'performdbupgradeto':
        raise NameError('Decorated function not a valid upgrader. '
                        'Must have name "perform_db_upgrade_N_to_M"')
    from_version = int(name_comps[3])
    to_version = int(name_comps[5])

    if not to_version == from_version+1:
        raise ValueError(f'Invalid upgrade versions in function name: '
                         f'{func.__name__}; upgrade from version '
                         f'{from_version} to version {to_version}.'
                         ' Can only upgrade from version N'
                         ' to version N+1')

    @wraps(func)
    def do_upgrade(conn: ConnectionPlus) -> None:

        log.info(f'Starting database upgrade version {from_version} '
                 f'to {to_version}')

        start_version = get_user_version(conn)
        if start_version != from_version:
            log.info(f'Skipping upgrade {from_version} -> {to_version} as'
                     f' current database version is {start_version}.')
            return

        # This function either raises or returns
        func(conn)

        set_user_version(conn, to_version)
        log.info(f'Succesfully performed upgrade {from_version} '
                 f'-> {to_version}')

    _UPGRADE_ACTIONS[to_version] = do_upgrade

    return do_upgrade


def perform_db_upgrade(conn: ConnectionPlus, version: int = -1) -> None:
    """
    This is intended to perform all upgrades as needed to bring the
    db from version 0 to the most current version (or the version specified).
    All the perform_db_upgrade_X_to_Y functions must raise if they cannot
    upgrade and be a NOOP if the current version is higher than their target.

    Args:
        conn: object for connection to the database
        version: Which version to upgrade to. We count from 0. -1 means
          'newest version'
    """
    version = _latest_available_version() if version == -1 else version

    current_version = get_user_version(conn)
    if current_version < version:
        log.info("Commencing database upgrade")
        for target_version in sorted(_UPGRADE_ACTIONS)[:version]:
            _UPGRADE_ACTIONS[target_version](conn)


# DATABASE UPGRADE FUNCTIONS


@upgrader
def perform_db_upgrade_0_to_1(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 0 to version 1

    Add a GUID column to the runs table and assign guids for all existing runs
    """

    sql = "SELECT name FROM sqlite_master WHERE type='table' AND name='runs'"
    cur = atomic_transaction(conn, sql)
    n_run_tables = len(cur.fetchall())

    if n_run_tables == 1:
        with atomic(conn) as conn:
            sql = "ALTER TABLE runs ADD COLUMN guid TEXT"
            transaction(conn, sql)
            # now assign GUIDs to existing runs
            cur = transaction(conn, 'SELECT run_id FROM runs')
            run_ids = [r[0] for r in many_many(cur, 'run_id')]

            pbar = tqdm(range(1, len(run_ids) + 1), file=sys.stdout)
            pbar.set_description("Upgrading database; v0 -> v1")

            for run_id in pbar:
                query = f"""
                        SELECT run_timestamp
                        FROM runs
                        WHERE run_id == {run_id}
                        """
                cur = transaction(conn, query)
                timestamp = one(cur, 'run_timestamp')
                timeint = int(np.round(timestamp*1000))
                sql = f"""
                        UPDATE runs
                        SET guid = ?
                        where run_id == {run_id}
                        """
                sampleint = 3736062718  # 'deafcafe'
                cur.execute(sql, (generate_guid(timeint=timeint,
                                                sampleint=sampleint),))
    else:
        raise RuntimeError(f"found {n_run_tables} runs tables expected 1")


@upgrader
def perform_db_upgrade_1_to_2(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 1 to version 2

    Add two indeces on the runs table, one for exp_id and one for GUID
    """

    sql = "SELECT name FROM sqlite_master WHERE type='table' AND name='runs'"
    cur = atomic_transaction(conn, sql)
    n_run_tables = len(cur.fetchall())

    pbar = tqdm(range(1), file=sys.stdout)
    pbar.set_description("Upgrading database; v1 -> v2")

    if n_run_tables == 1:
        _IX_runs_exp_id = """
                          CREATE INDEX
                          IF NOT EXISTS IX_runs_exp_id
                          ON runs (exp_id DESC)
                          """
        _IX_runs_guid = """
                        CREATE INDEX
                        IF NOT EXISTS IX_runs_guid
                        ON runs (guid DESC)
                        """
        with atomic(conn) as conn:
            # iterate through the pbar for the sake of the side effect; it
            # prints that the database is being upgraded
            for _ in pbar:
                transaction(conn, _IX_runs_exp_id)
                transaction(conn, _IX_runs_guid)
    else:
        raise RuntimeError(f"found {n_run_tables} runs tables expected 1")


@upgrader
def perform_db_upgrade_2_to_3(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 2 to version 3

    Insert a new column, run_description, to the runs table and fill it out
    for exisitng runs with information retrieved from the layouts and
    dependencies tables represented as the json output of a RunDescriber
    object
    """
    from qcodes.dataset.sqlite.db_upgrades.upgrade_2_to_3 import upgrade_2_to_3
    upgrade_2_to_3(conn)


@upgrader
def perform_db_upgrade_3_to_4(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 3 to version 4. This really
    repeats the version 3 upgrade as it originally had two bugs in
    the inferred annotation. inferred_from was passed incorrectly
    resulting in the parameter being marked inferred_from for each char
    in the inferred_from variable and inferred_from was not handled
    correctly for parameters that were neither dependencies nor dependent on
    other parameters. Both have since been fixed so rerun the upgrade.
    """
    from qcodes.dataset.sqlite.db_upgrades.upgrade_3_to_4 import upgrade_3_to_4
    upgrade_3_to_4(conn)


@upgrader
def perform_db_upgrade_4_to_5(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 4 to version 5.

    Make sure that 'snapshot' column always exists in the 'runs' table. This
    was not the case before because 'snapshot' was treated as 'metadata',
    hence the 'snapshot' column was dynamically created once there was a run
    with snapshot information.
    """
    with atomic(conn) as conn:
        pbar = tqdm(range(1), file=sys.stdout)
        pbar.set_description("Upgrading database; v4 -> v5")
        # iterate through the pbar for the sake of the side effect; it
        # prints that the database is being upgraded
        for _ in pbar:
            insert_column(conn, 'runs', 'snapshot', 'TEXT')


@upgrader
def perform_db_upgrade_5_to_6(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 5 to version 6.

    The upgrade ensures that the runs_description has a top-level entry
    called 'version'. Note that version changes of the runs_description will
    not be tracked as schema upgrades.
    """
    from qcodes.dataset.sqlite.db_upgrades.upgrade_5_to_6 import upgrade_5_to_6
    upgrade_5_to_6(conn)


@upgrader
def perform_db_upgrade_6_to_7(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 6 to version 7

    Add a captured_run_id and captured_counter column to the runs table and
    assign the value from the run_id and result_counter to these columns.
    """

    sql = "SELECT name FROM sqlite_master WHERE type='table' AND name='runs'"
    cur = atomic_transaction(conn, sql)
    n_run_tables = len(cur.fetchall())

    if n_run_tables == 1:

        pbar = tqdm(range(1), file=sys.stdout)
        pbar.set_description("Upgrading database; v6 -> v7")
        # iterate through the pbar for the sake of the side effect; it
        # prints that the database is being upgraded
        for _ in pbar:
            with atomic(conn) as conn:
                sql = "ALTER TABLE runs ADD COLUMN captured_run_id"
                transaction(conn, sql)
                sql = "ALTER TABLE runs ADD COLUMN captured_counter"
                transaction(conn, sql)

                sql = f"""
                        UPDATE runs
                        SET captured_run_id = run_id,
                            captured_counter = result_counter
                        """
                transaction(conn, sql)
    else:
        raise RuntimeError(f"found {n_run_tables} runs tables expected 1")


@upgrader
def perform_db_upgrade_7_to_8(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 7 to version 8.

    Add a new column to store the dataset's parents to the runs table.
    """
    with atomic(conn) as conn:
        pbar = tqdm(range(1), file=sys.stdout)
        pbar.set_description("Upgrading database; v7 -> v8")
        # iterate through the pbar for the sake of the side effect; it
        # prints that the database is being upgraded
        for _ in pbar:
            insert_column(conn, 'runs', 'parent_datasets', 'TEXT')


@upgrader
def perform_db_upgrade_8_to_9(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 8 to version 9.

    Add indices on the runs table for captured_run_id
    """

    sql = "SELECT name FROM sqlite_master WHERE type='table' AND name='runs'"
    cur = atomic_transaction(conn, sql)
    n_run_tables = len(cur.fetchall())

    pbar = tqdm(range(1), file=sys.stdout)
    pbar.set_description("Upgrading database; v8 -> v9")

    if n_run_tables == 1:
        _IX_runs_captured_run_id = """
                                CREATE INDEX
                                IF NOT EXISTS IX_runs_captured_run_id
                                ON runs (captured_run_id DESC)
                                """
        with atomic(conn) as connection:
            # iterate through the pbar for the sake of the side effect; it
            # prints that the database is being upgraded
            for _ in pbar:
                transaction(connection, _IX_runs_captured_run_id)
    else:
        raise RuntimeError(f"found {n_run_tables} runs tables expected 1")



from typing import Optional, Any

from qcodes.instrument.base import Instrument
from qcodes.instrument.ip import IPInstrument
# previous to introducing the `InstrumentLoggerAdapter` the IPToVisa instrument
# was logging in the name of the `VisaInstrument`. To maintain that behaviour
# import the `instrument.visa.log` and log to this one.
from qcodes.instrument.visa import VisaInstrument, VISA_LOGGER
from qcodes.instrument_drivers.american_magnetics.AMI430 import AMI430
from qcodes.utils.helpers import strip_attrs
from qcodes.logger.instrument_logger import get_instrument_logger
import qcodes.utils.validators as vals


# This module provides a class to make an IPInstrument behave like a
# VisaInstrument. This is only meant for use with the PyVISA-sim backend
# for testing purposes. If you really need an IPInstrument to become
# a VisaInstrument for actual instrument control, please rewrite the driver.

# At the end of the module, a 'zoo' of was-ip-is-now-visa drivers can be found.
# Such a driver is just a two-line class definition.


class IPToVisa(VisaInstrument, IPInstrument):  # type: ignore[misc]
    """
    Class to inject an VisaInstrument like behaviour in an
    IPInstrument that we'd like to use as a VISAInstrument with the
    simulation back-end.
    The idea is to inject this class just before the IPInstrument in
    the MRO. To avoid IPInstrument to ever take any effect, we sidestep
    it during the __init__ by calling directly to Instrument (which then
    class up through the chain) and explicitly reimplementing the __init__
    of VisaInstrument. We also must reimplement close, as that method
    calls super.

    Only meant for testing/simulation purposes!
    Do not use this for actual instrument control, there could be many
    nasty surprises.
    """

    def __init__(self, name: str, address: str,
                 port: Optional[int],
                 visalib: str,
                 device_clear: bool = False,
                 terminator: str = '\n',
                 timeout: float = 3,
                 **kwargs: Any):

        # remove IPInstrument-specific kwargs
        ipkwargs = ['write_confirmation']
        newkwargs = {kw: val for (kw, val) in kwargs.items()
                     if kw not in ipkwargs}

        Instrument.__init__(self, name, **newkwargs)
        self.visa_log = get_instrument_logger(self, VISA_LOGGER)

        ##################################################
        # __init__ of VisaInstrument

        self.add_parameter('timeout',
                           get_cmd=self._get_visa_timeout,
                           set_cmd=self._set_visa_timeout,
                           unit='s',
                           vals=vals.MultiType(vals.Numbers(min_value=0),
                                               vals.Enum(None)))

        # auxiliary VISA library to use for mocking
        self.visalib = visalib
        self.visabackend = ''

        self.set_address(address)
        if device_clear:
            self.device_clear()

        self.set_terminator(terminator)
        self.timeout.set(timeout)

    def close(self) -> None:
        """Disconnect and irreversibly tear down the instrument."""

        # VisaInstrument close
        if getattr(self, 'visa_handle', None):
            self.visa_handle.close()

        # Instrument close
        if hasattr(self, 'connection') and hasattr(self.connection, 'close'):
            self.connection.close()

        strip_attrs(self, whitelist=['name'])
        self.remove_instance(self)


class AMI430_VISA(AMI430, IPToVisa):  # type: ignore[misc]
    pass



# All manual references are to R&S RTO Digital Oscilloscope User Manual
# for firmware 3.65, 2017

import logging
import time
import warnings
from distutils.version import LooseVersion
from typing import Any, Optional

import numpy as np

from qcodes import Instrument
from qcodes.instrument.channel import InstrumentChannel
from qcodes.instrument.parameter import ArrayParameter
from qcodes.instrument.visa import VisaInstrument
from qcodes.utils import validators as vals
from qcodes.utils.helpers import create_on_off_val_mapping

log = logging.getLogger(__name__)


class ScopeTrace(ArrayParameter):
    def __init__(
        self, name: str, instrument: InstrumentChannel, channum: int, **kwargs: Any
    ) -> None:
        """
        The ScopeTrace parameter is attached to a channel of the oscilloscope.

        For now, we only support reading out the entire trace.
        """
        super().__init__(
            name=name,
            shape=(1,),
            label="Voltage",  # TODO: Is this sometimes dbm?
            unit="V",
            setpoint_names=("Time",),
            setpoint_labels=("Time",),
            setpoint_units=("s",),
            docstring="Holds scope trace",
            snapshot_value=False,
            instrument=instrument,
            **kwargs,
        )

        self.channel = instrument
        self.channum = channum
        self._trace_ready = False

    def prepare_trace(self) -> None:
        """
        Prepare the scope for returning data, calculate the setpoints
        """
        assert self.root_instrument is not None

        # We always use 16 bit integers for the data format
        self.root_instrument.dataformat("INT,16")
        # ensure little-endianess
        self.root_instrument.write("FORMat:BORder LSBFirst")
        # only export y-values
        self.root_instrument.write("EXPort:WAVeform:INCXvalues OFF")
        # only export one channel
        self.root_instrument.write("EXPort:WAVeform:MULTichannel OFF")

        # now get setpoints

        hdr = self.root_instrument.ask(f"CHANnel{self.channum}:" "DATA:HEADER?")
        hdr_vals = list(map(float, hdr.split(",")))
        t_start = hdr_vals[0]
        t_stop = hdr_vals[1]
        no_samples = int(hdr_vals[2])
        values_per_sample = hdr_vals[3]

        # NOTE (WilliamHPNielsen):
        # If samples are multi-valued, we need a `MultiParameter`
        # instead of an `ArrayParameter`.
        if values_per_sample > 1:
            raise NotImplementedError('There are several values per sample '
                                      'in this trace (are you using envelope'
                                      ' or peak detect?). We currently do '
                                      'not support saving such a trace.')

        self.shape = (no_samples,)
        self.setpoints = (tuple(np.linspace(t_start, t_stop, no_samples)),)

        self._trace_ready = True
        # we must ensure that all this took effect before proceeding
        self.root_instrument.ask("*OPC?")

    def get_raw(self) -> np.ndarray:
        """
        Returns a trace
        """

        instr = self.root_instrument
        assert instr is not None

        if not self._trace_ready:
            raise ValueError('Trace not ready! Please call '
                             'prepare_trace().')

        if instr.run_mode() == 'RUN Nx SINGLE':
            total_acquisitions = instr.num_acquisitions()
            completed_acquisitions = instr.completed_acquisitions()
            log.info(f'Acquiring {total_acquisitions} traces.')
            while completed_acquisitions < total_acquisitions:
                log.info(f'Acquired {completed_acquisitions}:'
                         f'{total_acquisitions}')
                time.sleep(0.25)
                completed_acquisitions = instr.completed_acquisitions()

        log.info('Acquisition completed. Polling trace from instrument.')
        vh = instr.visa_handle
        vh.write(f'CHANnel{self.channum}:DATA?')
        raw_vals = vh.read_raw()

        num_length = int(raw_vals[1:2])
        no_points = int(raw_vals[2:2+num_length])

        # cut of the header and the trailing '\n'
        raw_vals = raw_vals[2+num_length:-1]

        dataformat = instr.dataformat.get_latest()

        if dataformat == 'INT,8':
            int_vals = np.fromstring(raw_vals, dtype=np.int8, count=no_points)
        else:
            int_vals = np.fromstring(raw_vals, dtype=np.int16,
                                     count=no_points//2)

        # now the integer values must be converted to physical
        # values

        scale = self.channel.scale()
        no_divs = 10  # TODO: Is this ever NOT 10?

        # we always export as 16 bit integers
        quant_levels = 253*256
        conv_factor = scale*no_divs/quant_levels
        output = conv_factor*int_vals + self.channel.offset()

        return output


class ScopeMeasurement(InstrumentChannel):
    """
    Class to hold a measurement of the scope.
    """

    def __init__(self, parent: Instrument, name: str, meas_nr: int) -> None:
        """
        Args:
            parent: The instrument to which the channel is attached
            name: The name of the measurement
            meas_nr: The number of the measurement in question. Must match the
                actual number as used by the instrument (1..8)
        """

        if meas_nr not in range(1, 9):
            raise ValueError('Invalid measurement number; Min: 1, max 8')

        self.meas_nr = meas_nr
        super().__init__(parent, name)

        self.sources = vals.Enum('C1W1', 'C1W2', 'C1W3',
                                 'C2W1', 'C2W2', 'C2W3',
                                 'C3W1', 'C3W2', 'C3W3',
                                 'C4W1', 'C4W2', 'C4W3',
                                 'M1', 'M2', 'M3', 'M4',
                                 'R1', 'R2', 'R3',  'R4',
                                 'SBUS1', 'SBUS2', 'SBUS3', 'SBUS4',
                                 'D0', 'D1', 'D2', 'D3',
                                 'D4', 'D5', 'D6', 'D7',
                                 'D8', 'D9', 'D10', 'D11',
                                 'D12', 'D13', 'D14', 'D15',
                                 'TRK1', 'TRK2', 'TRK3', 'TRK4',
                                 'TRK5', 'TRK6', 'TRK7', 'TRK8',
                                 'SG1TL1', 'SG1TL2',
                                 'SG2TL1', 'SG2TL2',
                                 'SG3TL1', 'SG3TL2',
                                 'SG4TL1', 'SG4TL2',
                                 'Z1V1', 'Z1V2', 'Z1V3', 'Z1V4',
                                 'Z1I1', 'Z1I2', 'Z1I3', 'Z1I4',
                                 'Z2V1', 'Z2V2', 'Z2V3',  'Z2V4',
                                 'Z2I1', 'Z2I2', 'Z2I3', 'Z2I4')

        self.categories = vals.Enum('AMPTime', 'JITTer', 'EYEJitter',
                                    'SPECtrum', 'HISTogram', 'PROTocol')

        self.meas_type = vals.Enum(
                        # Amplitude/time measurements
                        'HIGH', 'LOW', 'AMPLitude',
                        'MAXimum', 'MINimum', 'PDELta',
                        'MEAN', 'RMS', 'STDDev',
                        'POVershoot', 'NOVershoot', 'AREA',
                        'RTIMe', 'FTIMe', 'PPULse',
                        'NPULse', 'PERiod', 'FREQuency',
                        'PDCYcle', 'NDCYcle', 'CYCarea',
                        'CYCMean', 'CYCRms', 'CYCStddev',
                        'PULCnt', 'DELay', 'PHASe',
                        'BWIDth', 'PSWitching', 'NSWitching',
                        'PULSetrain', 'EDGecount', 'SHT',
                        'SHR', 'DTOTrigger', 'PROBemeter',
                        'SLERising', 'SLEFalling',
                        # Jitter measurements
                        'CCJitter', 'NCJitter', 'CCWidth',
                        'CCDutycycle', 'TIE', 'UINTerval',
                        'DRATe', 'SKWDelay', 'SKWPhase',
                        # Eye diagram measurements
                        'ERPercent', 'ERDB', 'EHEight',
                        'EWIDth', 'ETOP', 'EBASe',
                        'QFACtor', 'RMSNoise', 'SNRatio',
                        'DCDistortion', 'ERTime', 'EFTime',
                        'EBRate', 'EAMPlitude', 'PPJitter',
                        'STDJitter', 'RMSJitter',
                        # Spectrum measurements
                        'CPOWer', 'OBWidth', 'SBWidth',
                        'THD', 'THDPCT', 'THDA',
                        'THDU', 'THDR', 'HAR',
                        'PLISt',
                        # Histogram measurements
                        'WCOunt', 'WSAMples', 'HSAMples',
                        'HPEak', 'PEAK', 'UPEakvalue',
                        'LPEakvalue', 'HMAXimum', 'HMINimum',
                        'MEDian', 'MAXMin', 'HMEan',
                        'HSTDdev', 'M1STddev', 'M2STddev',
                        'M3STddev', 'MKPositive', 'MKNegative'
                        )

        self.add_parameter('enable',
                           label=f'Measurement {meas_nr} enable',
                           set_cmd=f'MEASurement{meas_nr}:ENABle {{}}',
                           vals=vals.Enum('ON', 'OFF'),
                           docstring='Switches the measurement on or off.')

        self.add_parameter('source',
                           label=f'Measurement {meas_nr} source',
                           set_cmd=f'MEASurement{meas_nr}:SOURce {{}}',
                           vals=self.sources,
                           docstring='Set the source of a measurement if the '
                                     'measurement only needs one source.')

        self.add_parameter('source_first',
                           label=f'Measurement {meas_nr} first source',
                           set_cmd=f'MEASurement{meas_nr}:FSRC {{}}',
                           vals=self.sources,
                           docstring='Set the first source of a measurement'
                                     ' if the measurement only needs multiple'
                                     ' sources.')

        self.add_parameter('source_second',
                           label=f'Measurement {meas_nr} second source',
                           set_cmd=f'MEASurement{meas_nr}:SSRC {{}}',
                           vals=self.sources,
                           docstring='Set the second source of a measurement'
                                     ' if the measurement only needs multiple'
                                     ' sources.')

        self.add_parameter('category',
                           label=f'Measurement {meas_nr} category',
                           set_cmd=f'MEASurement{meas_nr}:CATegory {{}}',
                           vals=self.categories,
                           docstring='Set the category of a measurement.')

        self.add_parameter('main',
                           label=f'Measurement {meas_nr} main',
                           set_cmd=f'MEASurement{meas_nr}:MAIN {{}}',
                           vals=self.meas_type,
                           docstring='Set the main of a measurement.')

        self.add_parameter('statistics_enable',
                           label=f'Measurement {meas_nr} enable statistics',
                           set_cmd=f'MEASurement{meas_nr}:STATistics:ENABle'
                                   f' {{}}',
                           vals=vals.Enum('ON', 'OFF'),
                           docstring='Switches the measurement on or off.')

        self.add_parameter('clear',
                           label=f'Measurement {meas_nr} clear statistics',
                           set_cmd=f'MEASurement{meas_nr}:CLEar',
                           docstring='Clears/reset measurement.')

        self.add_parameter('event_count',
                           label=f'Measurement {meas_nr} number of events',
                           get_cmd=f'MEASurement{meas_nr}:RESult:EVTCount?',
                           get_parser=int,
                           docstring='Number of measurement results in the'
                                     ' long-term measurement.')

        self.add_parameter('result_avg',
                           label=f'Measurement {meas_nr} averages',
                           get_cmd=f'MEASurement{meas_nr}:RESult:AVG?',
                           get_parser=float,
                           docstring='Average of the long-term measurement'
                                     ' results.')


class ScopeChannel(InstrumentChannel):
    """
    Class to hold an input channel of the scope.

    Exposes: state, coupling, ground, scale, range, position, offset,
    invert, bandwidth, impedance, overload.
    """

    def __init__(self, parent: Instrument, name: str, channum: int) -> None:
        """
        Args:
            parent: The instrument to which the channel is attached
            name: The name of the channel
            channum: The number of the channel in question. Must match the
                actual number as used by the instrument (1..4)
        """

        if channum not in [1, 2, 3, 4]:
            raise ValueError('Invalid channel number! Must be 1, 2, 3, or 4.')

        self.channum = channum

        super().__init__(parent, name)

        self.add_parameter('state',
                           label=f'Channel {channum} state',
                           get_cmd=f'CHANnel{channum}:STATe?',
                           set_cmd=f'CHANnel{channum}:STATE {{}}',
                           vals=vals.Enum('ON', 'OFF'),
                           docstring='Switches the channel on or off')

        self.add_parameter('coupling',
                           label=f'Channel {channum} coupling',
                           get_cmd=f'CHANnel{channum}:COUPling?',
                           set_cmd=f'CHANnel{channum}:COUPling {{}}',
                           vals=vals.Enum('DC', 'DCLimit', 'AC'),
                           docstring=('Selects the connection of the channel'
                                      'signal. DC: 50 Ohm, DCLimit 1 MOhm, '
                                      'AC: Con. through DC capacitor'))

        self.add_parameter('ground',
                           label=f'Channel {channum} ground',
                           get_cmd=f'CHANnel{channum}:GND?',
                           set_cmd=f'CHANnel{channum}:GND {{}}',
                           vals=vals.Enum('ON', 'OFF'),
                           docstring=('Connects/disconnects the signal to/from'
                                      'the ground.'))

        # NB (WilliamHPNielsen): This parameter depends on other parameters and
        # should be dynamically updated accordingly. Cf. p 1178 of the manual
        self.add_parameter('scale',
                           label=f'Channel {channum} Y scale',
                           unit='V/div',
                           get_cmd=f'CHANnel{channum}:SCALe?',
                           set_cmd=self._set_scale,
                           get_parser=float,
                           )

        self.add_parameter('range',
                           label=f'Channel {channum} Y range',
                           unit='V',
                           get_cmd=f'CHANnel{channum}:RANGe?',
                           set_cmd=self._set_range,
                           get_parser=float
                           )

        # TODO (WilliamHPNielsen): would it be better to recast this in terms
        # of Volts?
        self.add_parameter('position',
                           label=f'Channel {channum} vert. pos.',
                           unit='div',
                           get_cmd=f'CHANnel{channum}:POSition?',
                           set_cmd=f'CHANnel{channum}:POSition {{}}',
                           get_parser=float,
                           vals=vals.Numbers(-5, 5),
                           docstring=('Positive values move the waveform up,'
                                      ' negative values move it down.'))

        self.add_parameter('offset',
                           label=f'Channel {channum} offset',
                           unit='V',
                           get_cmd=f'CHANnel{channum}:OFFSet?',
                           set_cmd=f'CHANnel{channum}:OFFSet {{}}',
                           get_parser=float,
                           )

        self.add_parameter('invert',
                           label=f'Channel {channum} inverted',
                           get_cmd=f'CHANnel{channum}:INVert?',
                           set_cmd=f'CHANnel{channum}:INVert {{}}',
                           vals=vals.Enum('ON', 'OFF'))

        # TODO (WilliamHPNielsen): This parameter should be dynamically
        # validated since 800 MHz BW is only available for 50 Ohm coupling
        self.add_parameter('bandwidth',
                           label=f'Channel {channum} bandwidth',
                           get_cmd=f'CHANnel{channum}:BANDwidth?',
                           set_cmd=f'CHANnel{channum}:BANDwidth {{}}',
                           vals=vals.Enum('FULL', 'B800', 'B200', 'B20')
                           )

        self.add_parameter('impedance',
                           label=f'Channel {channum} impedance',
                           unit='Ohm',
                           get_cmd=f'CHANnel{channum}:IMPedance?',
                           set_cmd=f'CHANnel{channum}:IMPedance {{}}',
                           vals=vals.Ints(1, 100000),
                           docstring=('Sets the impedance of the channel '
                                      'for power calculations and '
                                      'measurements.'))

        self.add_parameter('overload',
                           label=f'Channel {channum} overload',
                           get_cmd=f'CHANnel{channum}:OVERload?')

        self.add_parameter('arithmetics',
                           label=f'Channel {channum} arithmetics',
                           set_cmd=f'CHANnel{channum}:ARIThmetics {{}}',
                           get_cmd=f'CHANnel{channum}:ARIThmetics?',
                           val_mapping={'AVERAGE': 'AVER',
                                        'OFF': 'OFF',
                                        'ENVELOPE': 'ENV'}
                           )

        self.add_parameter('trace',
                           channum=self.channum,
                           parameter_class=ScopeTrace)

        self._trace_ready = False

    # Specialised/interlinked set/getters
    def _set_range(self, value: float) -> None:
        self.scale.cache.set(value/10)

        self._parent.write(f'CHANnel{self.channum}:RANGe {value}')

    def _set_scale(self, value: float) -> None:
        self.range.cache.set(value*10)

        self._parent.write(f'CHANnel{self.channum}:SCALe {value}')


class RTO1000(VisaInstrument):
    """
    QCoDeS Instrument driver for the
    Rohde-Schwarz RTO1000 series oscilloscopes.

    """

    def __init__(self, name: str, address: str,
                 model: Optional[str] = None, timeout: float = 5.,
                 HD: bool = True,
                 terminator: str = '\n',
                 **kwargs: Any) -> None:
        """
        Args:
            name: name of the instrument
            address: VISA resource address
            model: The instrument model. For newer firmware versions,
                this can be auto-detected
            timeout: The VISA query timeout
            HD: Does the unit have the High Definition Option (allowing
                16 bit vertical resolution)
            terminator: Command termination character to strip from VISA
                commands.
        """
        super().__init__(name=name, address=address, timeout=timeout,
                         terminator=terminator, **kwargs)

        # With firmware versions earlier than 3.65, it seems that the
        # model number can NOT be queried from the instrument
        # (at least fails with RTO1024, fw 2.52.1.1), so in that case
        # the user must provide the model manually.
        firmware_version = self.get_idn()['firmware']

        if LooseVersion(firmware_version) < LooseVersion('3'):
            log.warning('Old firmware version detected. This driver may '
                        'not be compatible. Please upgrade your firmware.')

        if LooseVersion(firmware_version) >= LooseVersion('3.65'):
            # strip just in case there is a newline character at the end
            self.model = self.ask('DIAGnostic:SERVice:WFAModel?').strip()
            if model is not None and model != self.model:
                warnings.warn("The model number provided by the user "
                              "does not match the instrument's response."
                              " I am going to assume that this oscilloscope "
                              f"is a model {self.model}")
        else:
            if model is None:
                raise ValueError('No model number provided. Please provide '
                                 'a model number (eg. "RTO1024").')
            else:
                self.model = model

        self.HD = HD

        # Now assign model-specific values
        self.num_chans = int(self.model[-1])
        self.num_meas = 8

        self._horisontal_divs = int(self.ask('TIMebase:DIVisions?'))

        self.add_parameter('display',
                           label='Display state',
                           set_cmd='SYSTem:DISPlay:UPDate {}',
                           val_mapping={'remote': 0,
                                        'view': 1})

        # Triggering

        self.add_parameter('trigger_display',
                           label='Trigger display state',
                           set_cmd='DISPlay:TRIGger:LINes {}',
                           get_cmd='DISPlay:TRIGger:LINes?',
                           val_mapping={'ON': 1, 'OFF': 0})

        # TODO: (WilliamHPNielsen) There are more available trigger
        # settings than implemented here. See p. 1261 of the manual
        # here we just use trigger1, which is the A-trigger

        self.add_parameter('trigger_source',
                           label='Trigger source',
                           set_cmd='TRIGger1:SOURce {}',
                           get_cmd='TRIGger1:SOURce?',
                           val_mapping={'CH1': 'CHAN1',
                                        'CH2': 'CHAN2',
                                        'CH3': 'CHAN3',
                                        'CH4': 'CHAN4',
                                        'EXT': 'EXT'})

        self.add_parameter('trigger_mode',
                           label='Trigger mode',
                           set_cmd='TRIGger:MODE {}',
                           get_cmd='TRIGger1:SOURce?',
                           vals=vals.Enum('AUTO', 'NORMAL', 'FREERUN'),
                           docstring='Sets the trigger mode which determines'
                                     ' the behaviour of the instrument if no'
                                     ' trigger occurs.\n'
                                     'Options: AUTO, NORMAL, FREERUN.',
                           unit='none')

        self.add_parameter('trigger_type',
                           label='Trigger type',
                           set_cmd='TRIGger1:TYPE {}',
                           get_cmd='TRIGger1:TYPE?',
                           val_mapping={'EDGE': 'EDGE',
                                        'GLITCH': 'GLIT',
                                        'WIDTH': 'WIDT',
                                        'RUNT': 'RUNT',
                                        'WINDOW': 'WIND',
                                        'TIMEOUT': 'TIM',
                                        'INTERVAL': 'INT',
                                        'SLEWRATE': 'SLEW',
                                        'DATATOCLOCK': 'DAT',
                                        'STATE': 'STAT',
                                        'PATTERN': 'PATT',
                                        'ANEDGE': 'ANED',
                                        'SERPATTERN': 'SERP',
                                        'NFC': 'NFC',
                                        'TV': 'TV',
                                        'CDR': 'CDR'}
                           )
        # See manual p. 1262 for an explanation of trigger types

        self.add_parameter('trigger_level',
                           label='Trigger level',
                           set_cmd=self._set_trigger_level,
                           get_cmd=self._get_trigger_level)

        self.add_parameter('trigger_edge_slope',
                           label='Edge trigger slope',
                           set_cmd='TRIGger1:EDGE:SLOPe {}',
                           get_cmd='TRIGger1:EDGE:SLOPe?',
                           vals=vals.Enum('POS', 'NEG', 'EITH'))

        # Horizontal settings

        self.add_parameter('timebase_scale',
                           label='Timebase scale',
                           set_cmd=self._set_timebase_scale,
                           get_cmd='TIMebase:SCALe?',
                           unit='s/div',
                           get_parser=float,
                           vals=vals.Numbers(25e-12, 10000))

        self.add_parameter('timebase_range',
                           label='Timebase range',
                           set_cmd=self._set_timebase_range,
                           get_cmd='TIMebase:RANGe?',
                           unit='s',
                           get_parser=float,
                           vals=vals.Numbers(250e-12, 100e3))

        self.add_parameter('timebase_position',
                           label='Horizontal position',
                           set_cmd=self._set_timebase_position,
                           get_cmd='TIMEbase:HORizontal:POSition?',
                           get_parser=float,
                           unit='s',
                           vals=vals.Numbers(-100e24, 100e24))

        # Acquisition

        # I couldn't find a way to query the run mode, so we manually keep
        # track of it. It is very important when getting the trace to make
        # sense of completed_acquisitions.
        self.add_parameter('run_mode',
                           label='Run/acquisition mode of the scope',
                           get_cmd=None,
                           set_cmd=None)

        self.run_mode('RUN CONT')

        self.add_parameter('num_acquisitions',
                           label='Number of single acquisitions to perform',
                           get_cmd='ACQuire:COUNt?',
                           set_cmd='ACQuire:COUNt {}',
                           vals=vals.Ints(1, 16777215),
                           get_parser=int)

        self.add_parameter('completed_acquisitions',
                           label='Number of completed acquisitions',
                           get_cmd='ACQuire:CURRent?',
                           get_parser=int)

        self.add_parameter('sampling_rate',
                           label='Sample rate',
                           docstring='Number of averages for measuring '
                           'trace.',
                           unit='Sa/s',
                           get_cmd='ACQuire:POINts:ARATe' + '?',
                           get_parser=int)

        self.add_parameter('acquisition_sample_rate',
                           label='Acquisition sample rate',
                           unit='Sa/s',
                           docstring='recorded waveform samples per second',
                           get_cmd='ACQuire:SRATe'+'?',
                           set_cmd='ACQuire:SRATe ' + ' {:.2f}',
                           vals=vals.Numbers(2, 20e12),
                           get_parser=float)

        # Data

        self.add_parameter('dataformat',
                           label='Export data format',
                           set_cmd='FORMat:DATA {}',
                           get_cmd='FORMat:DATA?',
                           vals=vals.Enum('ASC,0', 'REAL,32',
                                          'INT,8', 'INT,16'))

        # High definition mode (might not be available on all instruments)

        if HD:
            self.add_parameter('high_definition_state',
                               label='High definition (16 bit) state',
                               set_cmd=self._set_hd_mode,
                               get_cmd='HDEFinition:STAte?',
                               val_mapping=create_on_off_val_mapping(on_val=1,
                                                                     off_val=0),
                               docstring='Sets the filter bandwidth for the'
                                         ' high definition mode.\n'
                                         'ON: high definition mode, up to 16'
                                         ' bit digital resolution\n'
                                         'Options: ON, OFF\n\n'
                                         'Warning/Bug: By opening the HD '
                                         'acquisition menu on the scope, '
                                         'this value will be set to "ON".')

            self.add_parameter('high_definition_bandwidth',
                               label='High definition mode bandwidth',
                               set_cmd='HDEFinition:BWIDth {}',
                               get_cmd='HDEFinition:BWIDth?',
                               unit='Hz',
                               get_parser=float,
                               vals=vals.Numbers(1e4, 1e9))

        self.add_parameter('error_count',
                           label='Number of errors in the error stack',
                           get_cmd='SYSTem:ERRor:COUNt?',
                           unit='#',
                           get_parser=int)

        self.add_parameter('error_next',
                           label='Next error from the error stack',
                           get_cmd='SYSTem:ERRor:NEXT?',
                           get_parser=str)

        # Add the channels to the instrument
        for ch in range(1, self.num_chans+1):
            chan = ScopeChannel(self, f'channel{ch}', ch)
            self.add_submodule(f'ch{ch}', chan)

        for measId in range(1, self.num_meas+1):
            measCh = ScopeMeasurement(self, f'measurement{measId}', measId)
            self.add_submodule(f'meas{measId}', measCh)

        self.add_function('stop', call_cmd='STOP')
        self.add_function('reset', call_cmd='*RST')
        self.add_parameter('opc', get_cmd='*OPC?')
        self.add_parameter('stop_opc', get_cmd='STOP;*OPC?')
        self.add_parameter('status_operation',
                           get_cmd='STATus:OPERation:CONDition?',
                           get_parser=int)
        self.add_function('run_continues', call_cmd='RUNContinous')
        # starts the shutdown of the system
        self.add_function('system_shutdown', call_cmd='SYSTem:EXIT')

        self.connect_message()

    def run_cont(self) -> None:
        """
        Set the instrument in 'RUN CONT' mode
        """
        self.write('RUN')
        self.run_mode.set('RUN CONT')

    def run_single(self) -> None:
        """
        Set the instrument in 'RUN Nx SINGLE' mode
        """
        self.write('SINGLE')
        self.run_mode.set('RUN Nx SINGLE')

    def is_triggered(self) -> bool:
        wait_trigger_mask = 0b01000
        return bool(self.status_operation() & wait_trigger_mask) == False

    def is_running(self) -> bool:
        measuring_mask = 0b10000
        return bool(self.status_operation() & measuring_mask)

    def is_acquiring(self) -> bool:
        return self.is_triggered() & self.is_running()

    # Specialised set/get functions

    def _set_hd_mode(self, value: int) -> None:
        """
        Set/unset the high def mode
        """
        self._make_traces_not_ready()
        self.write(f'HDEFinition:STAte {value}')

    def _set_timebase_range(self, value: float) -> None:
        """
        Set the full range of the timebase
        """
        self._make_traces_not_ready()
        self.timebase_scale.cache.set(value/self._horisontal_divs)

        self.write(f'TIMebase:RANGe {value}')

    def _set_timebase_scale(self, value: float) -> None:
        """
        Set the length of one horizontal division.
        """
        self._make_traces_not_ready()
        self.timebase_range.cache.set(value*self._horisontal_divs)

        self.write(f'TIMebase:SCALe {value}')

    def _set_timebase_position(self, value: float) -> None:
        """
        Set the horizontal position.
        """
        self._make_traces_not_ready()
        self.write(f'TIMEbase:HORizontal:POSition {value}')

    def _make_traces_not_ready(self) -> None:
        """
        Make the scope traces be not ready.
        """
        self.ch1.trace._trace_ready = False
        self.ch2.trace._trace_ready = False
        self.ch3.trace._trace_ready = False
        self.ch4.trace._trace_ready = False

    def _set_trigger_level(self, value: float) -> None:
        """
        Set the trigger level on the currently used trigger source
        channel.
        """
        trans = {'CH1': 1, 'CH2': 2, 'CH3': 3, 'CH4': 4, 'EXT': 5}
        # We use get and not get_latest because we don't trust users to
        # not touch the front panel of an oscilloscope.
        source = trans[self.trigger_source.get()]
        if source != 5:
            submodule = self.submodules[f'ch{source}']
            assert isinstance(submodule, InstrumentChannel)
            v_range = submodule.range()
            offset = submodule.offset()

            if (value < -v_range/2 + offset) or (value > v_range/2 + offset):
                raise ValueError('Trigger level outside channel range.')

        self.write(f'TRIGger1:LEVel{source} {value}')

    def _get_trigger_level(self) -> float:
        """
        Get the trigger level from the currently used trigger source
        """
        trans = {'CH1': 1, 'CH2': 2, 'CH3': 3, 'CH4': 4, 'EXT': 5}
        # we use get and not get_latest because we don't trust users to
        # not touch the front panel of an oscilloscope
        source = trans[self.trigger_source.get()]

        val = self.ask(f'TRIGger1:LEVel{source}?')

        return float(val.strip())
